import functools
import logging


def update_current_tenant(function):

    @functools.wraps(function)
    def call_and_update_current_tenant(self, *args):
        ret = None
        path = args[0]
        if path == '/currentPortal':
            tenant = args[1]
            logging.getLogger().debug('Updating current tenant. %s', {'tenant': tenant})
            session = self.session()
            if not session.is_local_auth():  # Skip calling the function if using local authentication
                ret = function(self, *args)
            self.session().update_tenant(tenant)
            logging.getLogger().debug('Updated current tenant. %s', {'tenant': tenant})
        else:
            ret = function(self, *args)
        return ret

    return call_and_update_current_tenant

# ValidIpAddressRegex = "^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$";

# ValidHostnameRegex = "^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$";



import logging

from ..common import Object
from ..exception import CTERAException
from .base_command import BaseCommand


class Array(BaseCommand):
    """ Gateway Array APIs """

    def get(self, name=None):
        """
        Get Array. If an array name was not passed as an argument, a list of all arrays will be retrieved

        :param str,optional name: Name of the array
        """
        return self._gateway.get('/config/storage/arrays' + ('' if name is None else ('/' + name)))

    def add(self, array_name, level, members=None):
        """
        Add a new array

        :param str array_name: Name for the new array
        :param RAIDLevel level: RAID level
        :param list(str) members: Members of the array. If not specified, the system will try to create an array using all available drives
        """
        param = Object()
        param.name = array_name
        param.level = level
        param.members = [drive.name for drive in self._gateway.drive.get_status()] if members is None else members

        try:
            logging.getLogger().info("Creating a storage array.")
            response = self._gateway.add("/config/storage/arrays", param)
            logging.getLogger().info("Storage array created.")
            return response
        except CTERAException as error:
            logging.getLogger().error("Storage array creation failed.")
            raise CTERAException("Storage array creation failed.", error)

    def delete(self, array_name):
        """
        Delete an array

        :param str name: The name of the array to delete
        """
        try:
            logging.getLogger().info("Deleting a storage array.")
            response = self._gateway.delete("/config/storage/arrays/" + array_name)
            logging.getLogger().info("Storage array deleted. %s", {'array_name': array_name})
            return response
        except CTERAException as error:
            logging.getLogger().error("Storage array deletion failed.")
            raise CTERAException("Storage array deletion failed.", error)

    def delete_all(self):
        """
        Delete all arrays
        """
        arrays = self._gateway.get('/config/storage/arrays')
        for array in arrays:
            self.delete(array.name)



import logging

from ..exception import CTERAException
from .base_command import BaseCommand


class Login(BaseCommand):

    def info(self):
        """
        Get login info
        """
        return self._gateway.get('/nosession/logininfo')

    def login(self, username, password):
        host = self._gateway.host()
        try:
            self._gateway.form_data('/login', {'username': username, 'password': password})
            logging.getLogger().info("User logged in. %s", {'host': host, 'user': username})
            self._gateway.mtool.login()
        except CTERAException as error:
            logging.getLogger().error("Login failed. %s", {'host': host, 'user': username})
            raise error

    def logout(self):
        host = self._gateway.host()
        user = self._gateway.session().user.name
        try:
            self._gateway.form_data('/logout', {'foo': 'bar'})
            logging.getLogger().info("User logged out. %s", {'host': host, 'user': user})
        except CTERAException as error:
            logging.getLogger().error("Logout failed. %s", {'host': host, 'user': user})
            raise error



from ...lib import Iterator, Command
from .fetch_resources_param import FetchResourcesParamBuilder


def list_dir(ctera_host, param):
    response = fetch_resources(ctera_host, param)
    return (response.hasMore, response.items)


def fetch_resources(ctera_host, param):
    return ctera_host.execute('', 'fetchResources', param)


def ls(ctera_host, path, depth=1, include_deleted=False):
    builder = FetchResourcesParamBuilder().root(path.encoded_fullpath()).depth(depth)
    if include_deleted:
        builder.include_deleted()
    param = builder.build()
    if depth > 0:
        function = Command(list_dir, ctera_host)
        return Iterator(function, param)
    return fetch_resources(ctera_host, param)



import logging

from . import enum
from .files import path
from ..common import Object
from ..exception import CTERAException, InputError
from .base_command import BaseCommand
from .types import NFSv3AccessControlEntry, RemoveNFSv3AccessControlEntry, ShareAccessControlEntry, RemoveShareAccessControlEntry


class Shares(BaseCommand):

    def get(self, name=None):
        """
        Get Share. If a share name was not passed as an argument, a list of all shares will be retrieved
        :param str,optional name: Name of the share
        """
        return self._gateway.get('/config/fileservices/share' + ('' if name is None else ('/' + name)))

    def add(self,
            name,
            directory,
            acl=None,
            access=enum.Acl.WindowsNT,
            csc=enum.ClientSideCaching.Manual,
            dir_permissions=777,
            comment=None,
            export_to_afp=False,
            export_to_ftp=False,
            export_to_nfs=False,
            export_to_pc_agent=False,
            export_to_rsync=False,
            indexed=False,
            trusted_nfs_clients=None,
            uuid=None
            ):  # pylint: disable=too-many-arguments,too-many-locals,unused-argument
        """
        Add a network share.

        :param str name: The share name
        :param str directory: Full directory path
        :param list[cterasdk.edge.types.ShareAccessControlEntry] acl: List of access control entries
        :param cterasdk.edge.enum.Acl access: The Windows File Sharing authentication mode, defaults to ``winAclMode``
        :param cterasdk.edge.enum.ClientSideCaching csc: The client side caching (offline files) configuration, defaults to ``manual``
        :param int dir_permissions: Directory Permission, defaults to 777
        :param str comment: Comment
        :param bool export_to_afp: Whether to enable AFP access, defaults to ``False``
        :param bool export_to_ftp: Whether to enable FTP access, defaults to ``False``
        :param bool export_to_nfs: Whether to enable NFS access, defaults to ``False``
        :param bool export_to_pc_agent: Whether to allow as a destination share for CTERA Backup Agents, defaults to ``False``
        :param bool export_to_rsync: Whether to enable access over rsync, defaults to ``False``
        :param bool indexed: Whether to enable indexing for search, defaults to ``False``
        :param list[cterasdk.edge.types.NFSv3AccessControlEntry] trusted_nfs_clients: Trusted NFS v3 clients, defaults to ``None``
        """
        acl = acl or []

        param = Object()
        param.name = name

        parts = path.CTERAPath(directory, '/').parts()
        volume = parts[0]
        self._validate_root_directory(volume)
        param.volume = volume

        directory = '/'.join(parts[1:])
        param.directory = directory

        param.access = access
        param.clientSideCaching = csc
        param.dirPermissions = dir_permissions
        param.exportToAFP = export_to_afp
        param.exportToFTP = export_to_ftp
        param.exportToNFS = export_to_nfs
        param.exportToPCAgent = export_to_pc_agent
        param.exportToRSync = export_to_rsync
        param.indexed = indexed
        param.comment = comment
        Shares._validate_acl(acl)
        param.acl = [acl_entry.to_server_object() for acl_entry in acl]
        param.trustedNFSClients = [client.to_server_object() for client in (trusted_nfs_clients or [])]
        if uuid:
            param._uuid = uuid  # pylint: disable=protected-access

        try:
            self._gateway.add('/config/fileservices/share', param)
            logging.getLogger().info("Share created. %s", {'name': name})
        except Exception as error:
            logging.getLogger().error("Share creation failed.")
            raise CTERAException('Share creation failed', error)

    def set_share_winacls(self, name):
        """
        Set a network share to use Windows ACL Emulation Mode

        :param str name: The share name
        """
        logging.getLogger().error("Updating Windows file sharing access mode. %s", {'share': name, 'access': enum.Acl.WindowsNT})
        self._gateway.put('/config/fileservices/share/' + name + '/access', enum.Acl.WindowsNT)

    def get_access_type(self, name):
        """
        Get the network share Windows File Sharing authentication mode

        :param str name: The share name
        """
        return self._gateway.get('/config/fileservices/share/' + name + '/access')

    def set_access_type(self, name, access):
        """
        Set the network share Windows File Sharing authentication mode

        :param str name: The share name
        :param cterasdk.edge.enum.Acl access: The Windows File Sharing authentication mode
        """
        logging.getLogger().info("Updating Windows file sharing access mode. %s", {'share': name, 'access': access})
        self._gateway.put('/config/fileservices/share/' + name + '/access', access)

    def block_files(self, name, extensions):
        """
        Configure a share to block one or more file extensions

        :param str name: The share name
        :param list[str] extensions: List of file extensions to block
        """
        share = self.get(name)
        if share.access != enum.Acl.WindowsNT:
            raise CTERAException('Cannot block file types on non Windows-ACL enabled shares', None, share=share.name, access=share.access)
        logging.getLogger().error("Updating the list of blocked file extensions. %s",
                                  {'share': name, 'extensions': extensions, 'access': enum.Acl.WindowsNT})
        self._gateway.put('/config/fileservices/share/' + share.name + '/screenedFileTypes', extensions)

    def set_acl(self, name, acl):
        """
        Set a network share's access control entries.

        :param str name: The share name
        :param list[cterasdk.edge.types.ShareAccessControlEntry] acl: List of access control entries

        .. warning:: this method will override the existing access control entries
        """
        Shares._validate_acl(acl)

        param = [acl_entry.to_server_object() for acl_entry in acl]
        self._gateway.put('/config/fileservices/share/' + name + '/acl', param)

    def add_acl(self, name, acl):
        """
        Add one or more access control entries to an existing share.

        :param str name: The share name
        :param list[cterasdk.edge.types.ShareAccessControlEntry] acl: List of access control entries to add
        """
        Shares._validate_acl(acl)

        current_acl = self.get_acl(name)

        new_acl_dict = {
            acl_entry.principal_type + '#' + acl_entry.name: acl_entry.to_server_object()
            for acl_entry in acl
        }

        for entry in current_acl:
            ace = ShareAccessControlEntry.from_server_object(entry)
            entry_key = ace.principal_type + '#' + ace.name
            if entry_key not in new_acl_dict:
                new_acl_dict[entry_key] = entry

        acls_array = [v for k, v in new_acl_dict.items()]
        self._gateway.put('/config/fileservices/share/' + name + '/acl', acls_array)

    def remove_acl(self, name, acl):
        """
        Remove one or more access control entries from an existing share.

        :param str name: The share name
        :param list[cterasdk.edge.types.RemoveShareAccessControlEntry] acl: List of access control entries to remove
        """
        Shares._validate_remove_acl(acl)

        remove_acl_dict = {
            acl_entry.principal_type + '#' + acl_entry.name: True
            for acl_entry in acl
        }

        current_acl = self.get_acl(name)

        new_acl = []
        for entry in current_acl:
            ace = ShareAccessControlEntry.from_server_object(entry)
            if not remove_acl_dict.get(ace.principal_type + '#' + ace.name, False):
                new_acl.append(entry)

        self._gateway.put('/config/fileservices/share/' + name + '/acl', new_acl)

    def get_acl(self, name):
        """
        Get the current access control entries from an existing share.

        :param str name: The share name
        """
        return self._gateway.get('/config/fileservices/share/' + name + '/acl')

    def modify(
            self,
            name,
            directory=None,
            acl=None,
            access=None,
            csc=None,
            dir_permissions=None,
            comment=None,
            export_to_afp=None,
            export_to_ftp=None,
            export_to_nfs=None,
            export_to_pc_agent=None,
            export_to_rsync=None,
            indexed=None,
            trusted_nfs_clients=None
                ):  # pylint: disable=too-many-arguments,too-many-locals,too-many-branches,unused-argument
        """
        Modify an existing network share. All parameters but name are optional and default to None

        :param str name: The share name
        :param str,optional directory: Full directory path
        :param list[cterasdk.edge.types.ShareAccessControlEntry],optional acl: List of access control entries
        :param cterasdk.edge.enum.Acl,optional access: The Windows File Sharing authentication mode
        :param cterasdk.edge.enum.ClientSideCaching,optional csc: The client side caching (offline files) configuration
        :param int,optional dir_permissions: Directory Permission
        :param str,optional comment: Comment
        :param bool,optional export_to_afp: Whether to enable AFP access
        :param bool,optional export_to_ftp: Whether to enable FTP access
        :param bool,optional export_to_nfs: Whether to enable NFS access
        :param bool,optional export_to_pc_agent: Whether to allow as a destination share for CTERA Backup Agents
        :param bool,optional export_to_rsync: Whether to enable access over rsync
        :param bool,optional indexed: Whether to enable indexing for search
        :param list[cterasdk.edge.types.NFSv3AccessControlEntry] trusted_nfs_clients: Trusted NFS v3 clients, defaults to ``None``
        """
        share = self.get(name=name)
        if directory is not None:
            parts = path.CTERAPath(directory, '/').parts()
            volume = parts[0]
            self._validate_root_directory(volume)
            share.volume = volume
            directory = '/'.join(parts[1:])
            share.directory = directory
        if access is not None:
            share.access = access
        if csc is not None:
            share.clientSideCaching = csc
        if dir_permissions is not None:
            share.dirPermissions = dir_permissions
        if export_to_afp is not None:
            share.exportToAFP = export_to_afp
        if export_to_ftp is not None:
            share.exportToFTP = export_to_ftp
        if export_to_nfs is not None:
            share.exportToNFS = export_to_nfs
        if export_to_pc_agent is not None:
            share.exportToPCAgent = export_to_pc_agent
        if export_to_rsync is not None:
            share.exportToRSync = export_to_rsync
        if indexed is not None:
            share.indexed = indexed
        if comment is not None:
            share.comment = comment
        if acl is not None:
            Shares._validate_acl(acl)
            share.acl = [acl_entry.to_server_object() for acl_entry in acl]
        if trusted_nfs_clients is not None:
            share.trustedNFSClients = [client.to_server_object() for client in trusted_nfs_clients]

        try:
            self._gateway.put('/config/fileservices/share/' + name, share)
            logging.getLogger().info("Share modified. %s", {'name': name})
        except Exception as error:
            msg = f'Failed to modify the share {name}'
            logging.getLogger().error(msg)
            raise CTERAException(msg, error)

    def delete(self, name):
        """
        Delete a share.

        :param str name: The share name
        """
        try:
            self._gateway.delete('/config/fileservices/share/' + name)
            logging.getLogger().info("Share deleted. %s", {'name': name})
        except Exception as error:
            logging.getLogger().error("Share deletion failed.")
            raise CTERAException('Share deletion failed', error)

    def get_trusted_nfs_clients(self, name):
        """
        Get the current trusted NFS client entries from an existing share.

        :param str name: The share name
        """
        return self._gateway.get('/config/fileservices/share/' + name + '/trustedNFSClients')

    def set_trusted_nfs_clients(self, name, trusted_nfs_clients):
        """
        Set a network share's trusted NFS client entries.

        :param str name: The share name
        :param list[cterasdk.edge.types.NFSv3AccessControlEntry] trusted_nfs_clients: Trusted NFS v3 clients

        .. warning:: this method will override the existing access control entries
        """
        Shares._validate_trusted_nfs_clients(trusted_nfs_clients)

        param = [client.to_server_object() for client in (trusted_nfs_clients or [])]
        self._gateway.put('/config/fileservices/share/' + name + '/trustedNFSClients', param)

    def add_trusted_nfs_clients(self, name, trusted_nfs_clients):
        """
        Add one or more trusted NFS client entries to an existing share.

        :param str name: The share name
        :param list[cterasdk.edge.types.NFSv3AccessControlEntry] trusted_nfs_clients: Trusted NFS v3 clients
        """
        Shares._validate_trusted_nfs_clients(trusted_nfs_clients)

        new_trusted_nfs_clients_dict = {
            trusted_nfs_client.address + '#' + trusted_nfs_client.netmask: trusted_nfs_client.to_server_object()
            for trusted_nfs_client in trusted_nfs_clients
        }

        def entry_not_in_new(entry):
            trusted_nfs_client = NFSv3AccessControlEntry.from_server_object(entry)
            entry_key = trusted_nfs_client.address + '#' + trusted_nfs_client.netmask
            return entry_key not in new_trusted_nfs_clients_dict

        param = list(new_trusted_nfs_clients_dict.values()) + list(filter(entry_not_in_new, self.get_trusted_nfs_clients(name)))

        self._gateway.put('/config/fileservices/share/' + name + '/trustedNFSClients', param)

    def remove_trusted_nfs_clients(self, name, trusted_nfs_clients):
        """
        Remove one or more trusted NFS client entries from an existing share.

        :param str name: The share name
        :param list[cterasdk.edge.types.RemoveNFSv3AccessControlEntry] trusted_nfs_clients: Trusted NFS v3 clients
        """
        Shares._validate_remove_trusted_nfs_clients(trusted_nfs_clients)

        remove_trusted_nfs_clients_dict = {
            trusted_nfs_client.address + '#' + trusted_nfs_client.netmask
            for trusted_nfs_client in trusted_nfs_clients
        }

        def entry_not_removed(entry):
            trusted_nfs_client = NFSv3AccessControlEntry.from_server_object(entry)
            entry_key = trusted_nfs_client.address + '#' + trusted_nfs_client.netmask
            return entry_key not in remove_trusted_nfs_clients_dict

        param = list(filter(entry_not_removed, self.get_trusted_nfs_clients(name)))
        self._gateway.put('/config/fileservices/share/' + name + '/trustedNFSClients', param)

    def get_screened_file_types(self, name):
        """
        Get the share's current list of blocked file extensions

        :param str name: The share name
        """
        return self._gateway.get('/config/fileservices/share/' + name + '/screenedFileTypes')

    def set_screened_file_types(self, name, extensions):
        """
        Set the share's current list of blocked file extensions (override the current list)

        :param str name: The share name
        :param list[str] extensions: List of file extensions to block
        """
        share = self.get(name)
        if share.access != enum.Acl.WindowsNT:
            raise CTERAException('Cannot block file types on non Windows-ACL enabled shares', None, share=share.name, access=share.access)
        logging.getLogger().info(
            "Updating the list of blocked file extensions. %s",
            {'share': name, 'extensions': extensions, 'access': enum.Acl.WindowsNT}
        )
        self._gateway.put('/config/fileservices/share/' + share.name + '/screenedFileTypes', extensions)

    def add_screened_file_types(self, name, extensions):
        """
        Add extensions to the share's current list of blocked file extensions

        :param str name: The share name
        :param list[str] extensions: List of file extensions to add
        """
        share = self.get(name)
        if share.access != enum.Acl.WindowsNT:
            raise CTERAException('Cannot block file types on non Windows-ACL enabled shares', None, share=share.name, access=share.access)

        new_list = list(set(share.screenedFileTypes + extensions))

        logging.getLogger().info(
            "Updating the list of blocked file extensions. %s",
            {'share': name, 'extensions': new_list, 'access': enum.Acl.WindowsNT}
        )
        self._gateway.put('/config/fileservices/share/' + share.name + '/screenedFileTypes', new_list)

    def remove_screened_file_types(self, name, extensions):
        """
        Remove extensions from the share's current list of blocked file extensions

        :param str name: The share name
        :param list[str] extensions: List of file extensions to remove
        """
        share = self.get(name)
        if share.access != enum.Acl.WindowsNT:
            raise CTERAException('Cannot block file types on non Windows-ACL enabled shares', None, share=share.name, access=share.access)

        new_list = list(set(share.screenedFileTypes) - set(extensions))

        logging.getLogger().info(
            "Updating the list of blocked file extensions. %s",
            {'share': name, 'extensions': new_list, 'access': enum.Acl.WindowsNT}
        )
        self._gateway.put('/config/fileservices/share/' + share.name + '/screenedFileTypes', new_list)

    def _validate_root_directory(self, name):
        param = Object()
        param.path = '/'

        response = self._gateway.execute('/status/fileManager', 'listPhysicalFolders', param)
        for root in response:
            if root.fullpath == f'/{name}':
                logging.getLogger().debug("Found root directory. %s", {'name': root.name, 'type': root.type, 'fullpath': root.fullpath})
                return name

        logging.getLogger().error("Could not find root directory. %s", {'name': name})

        options = [root.fullpath[1:] for root in response]
        raise InputError('Invalid root directory.', name, options)

    @staticmethod
    def _validate_acl(acl):
        if not isinstance(acl, list):
            raise InputError('Invalid access control list format', repr(acl), '[cterasdk.edge.types.ShareAccessControlEntry, ...]')
        for acl_entry in acl:
            if not isinstance(acl_entry, ShareAccessControlEntry):
                raise InputError('Invalid access control entry format', repr(acl_entry), 'cterasdk.edge.types.ShareAccessControlEntry')

    @staticmethod
    def _validate_remove_acl(acl):
        if not isinstance(acl, list):
            raise InputError('Invalid access control list format', repr(acl), '[cterasdk.edge.types.RemoveShareAccessControlEntry, ...]')
        for acl_entry in acl:
            if not isinstance(acl_entry, RemoveShareAccessControlEntry):
                raise InputError(
                    'Invalid access control entry format',
                    repr(acl_entry),
                    'cterasdk.edge.types.RemoveShareAccessControlEntry'
                )

    @staticmethod
    def _validate_trusted_nfs_clients(trusted_nfs_clients):
        if not isinstance(trusted_nfs_clients, list):
            raise InputError(
                'Invalid Trusted NFS Clients list format',
                repr(trusted_nfs_clients),
                '[cterasdk.edge.types.NFSv3AccessControlEntry, ...]'
            )
        for entry in trusted_nfs_clients:
            if not isinstance(entry, NFSv3AccessControlEntry):
                raise InputError('Invalid Trusted NFS Clients entry format', repr(entry), 'cterasdk.edge.types.NFSv3AccessControlEntry')

    @staticmethod
    def _validate_remove_trusted_nfs_clients(trusted_nfs_clients):
        if not isinstance(trusted_nfs_clients, list):
            raise InputError(
                'Invalid Trusted NFS Clients list format',
                repr(trusted_nfs_clients),
                '[cterasdk.edge.types.RemoveNFSv3AccessControlEntry, ...]'
            )
        for entry in trusted_nfs_clients:
            if not isinstance(entry, RemoveNFSv3AccessControlEntry):
                raise InputError(
                    'Invalid Trusted NFS Clients entry format',
                    repr(entry),
                    'cterasdk.edge.types.RemoveNFSv3AccessControlEntry'
                )



import logging

from ...common import Object
from ...exception import CTERAException
from .path import CTERAPath


class ItemExists(CTERAException):
    pass


class InvalidPath(CTERAException):
    pass


class InvalidName(CTERAException):
    pass


class ReservedName(CTERAException):
    pass


FileManagerTaskRC = {
    "FileWithTheSameNameExist": ItemExists,
    "DestinationNotExists": InvalidPath,
    "InvalidName": InvalidName,
    "ReservedName": ReservedName
}


def mkdir(ctera_host, path, recurse=False):
    if recurse:
        array = path.parts()
        for i in range(1, len(array) + 1):
            dirpath = CTERAPath('/'.join(array[:i]), path.basepath)
            try:
                _mkdir(ctera_host, dirpath)
            except ItemExists:
                pass
    else:
        _mkdir(ctera_host, path)


def _mkdir(ctera_host, path):
    param = Object()
    param.name = path.name()
    param.parentPath = path.parent().encoded_fullpath()

    relativepath = str(path.relativepath)

    logging.getLogger().info('Creating directory. %s', {'path': relativepath})

    response = ctera_host.execute('', 'makeCollection', param)
    _process_response(response, relativepath)


def _process_response(response, path):
    try:
        _process_error(response, path)
    except ItemExists as error:
        logging.getLogger().warning('A file or folder with the same name already exists. %s', {'path': path})
        raise error
    except InvalidPath as error:
        logging.getLogger().error('Invalid parent directory path. %s', {'path': path})
        raise error
    except InvalidName as error:
        logging.getLogger().error('Directory name contains invalid characters. %s', {'name': path})
        raise error
    except ReservedName as error:
        logging.getLogger().error('Reserved directory name. %s', {'name': path})
        raise error
    logging.getLogger().info('Directory created. %s', {'path': path})


def _process_error(response, path):
    error = FileManagerTaskRC.get(response)
    if error is not None:
        error = error()
        error.path = path
        raise error



import logging

from .common import SrcDstParam, ActionResourcesParam


def rename(ctera_host, path, name):
    rename_param = ActionResourcesParam.instance()
    rename_param.add(SrcDstParam.instance(src=path.fullpath(), dest=path.parent().joinpath(name).fullpath()))

    logging.getLogger().info('Renaming item. %s', {'path': str(path.relativepath), 'name': name})

    return ctera_host.execute('', 'moveResources', rename_param)



import logging

from ..exception import ObjectNotFoundException
from .base_command import BaseCommand
from ..lib import Iterator, Command
from ..common import Object
from ..common import union
from . import enum
from . import query


class Portals(BaseCommand):
    """
    Global Admin Portals APIs
    """

    default = ['name']

    def get(self, name, include=None):
        """
        Get a tenant

        :param str name: Name of the tenant
        :param list[str] include: List of fields to retrieve, defaults to ['name']
        """
        include = union(include or [], Portals.default)
        include = ['/' + attr for attr in include]
        tenant = self._portal.get_multi('/portals/' + name, include)
        if tenant.name is None:
            raise ObjectNotFoundException('Could not find tenant', f'/portals/{name}', name=name)
        return tenant

    def list_tenants(self, include=None, portal_type=None):
        """
        List tenants.\n
        To retrieve tenants, you must first browse the Global Administration Portal, using: `GlobalAdmin.portals.browse_global_admin()`

        :param list[str],optional include: List of fields to retrieve, defaults to ['name']
        :param cterasdk.core.enum.PortalType portal_type: The Portal type
        """
        # browse administration
        include = union(include or [], Portals.default)
        param = query.QueryParamBuilder().include(include).build()
        baseurl = '/portals'
        if portal_type == enum.PortalType.Team:
            baseurl = '/teamPortals'
        elif portal_type == enum.PortalType.Reseller:
            baseurl = '/resellerPortals'
        return query.iterator(self._portal, baseurl, param)

    def _query_portals(self, param):
        response = self._portal.execute('', 'getPortalsDisplayInfo', param)
        return (response.hasMore, response.objects)

    def tenants(self, include_deleted=False):
        """
        Get all tenants

        :param bool,optional include_deleted: Include deleted tenants, defaults to False
        """
        # Check if current session is global admin
        param = query.QueryParamBuilder().include_classname().put('isTrashcan', include_deleted).build()
        function = Command(self._query_portals)
        return Iterator(function, param)

    def add(self, name, display_name=None, billing_id=None, company=None, plan=None, comment=None):
        """
        Add a new tenant

        :param str name: Name of the new tenant
        :param str,optional display_name: Display Name of the new tenant, defaults to None
        :param str,optional billing_id: Billing ID of the new tenant, defaults to None
        :param str,optional company: Company Name of the new tenant, defaults to None
        :param str,optional plan: Subscription plan name to assign to the new tenant, defaults to None
        :param str,optional comment: Assign a comment to the new tenant, defaults to None
        :return str: A relative url path to the Team Portal
        """

        param = Object()
        if plan:
            param.plan = self._portal.plans.get(plan, include=['baseObjectRef']).baseObjectRef
        param._classname = 'TeamPortal'  # pylint: disable=protected-access
        param.name = name
        param.displayName = display_name
        param.externalPortalId = billing_id
        param.companyName = company
        param.comment = comment

        logging.getLogger().info('Creating Team Portal. %s', {'name': name})

        response = self._portal.add('/teamPortals', param)

        logging.getLogger().info('Team Portal created. %s', {'name': name})

        return response

    def subscribe(self, tenant, plan):
        """
        Subscribe a tenant to a plan

        :param str name: Name of the tenant
        :param str,plan: Name of the subscription plan
        """
        return self._portal.execute('/portals/' + tenant, 'subscribe', plan)

    def delete(self, name):
        """
        Delete an existing tenant

        :param str name: Name of the tenant to delete
        """
        logging.getLogger().info('Deleting Portal. %s', {'name': name})

        response = self._portal.execute('/teamPortals/' + name, 'delete')

        logging.getLogger().info('Portal deleted. %s', {'name': name})

        return response

    def undelete(self, name):
        """
        Undelete a previously deleted tenant

        :param str name: Name of the tenant to undelete
        """
        logging.getLogger().info('Recovering Portal. %s', {'name': name})

        response = self._portal.execute('/teamPortals/' + name, 'moveFromTrashcan')

        logging.getLogger().info('Portal recovered. %s', {'name': name})

        return response

    def browse(self, tenant):
        """
        Browse a tenant

        :param str tenant: Name of the tenant to browse
        """
        self._portal.put('/currentPortal', tenant)

    def browse_global_admin(self):
        """
        Browse the Global Admin
        """
        self.browse('')

    def apply_changes(self, wait=False):
        """
        Apply provisioning changes.\n

        :param bool,optional wait: Wait for all changes to apply
        """
        param = Object()
        param.objectId = None
        param.type = 'portals'
        logging.getLogger().info('Applying provisioning changes.')
        task = self._portal.execute('', 'updatePortals', param)
        if wait:
            task = self._portal.tasks.wait(task)
        return task



import logging
from ...common import DateTimeUtils

from .common import CreateShareParam


def mklink(ctera_host, path, access, expire_in):
    access = {'RO': 'ReadOnly', 'RW': 'ReadWrite', 'PO': 'PreviewOnly'}.get(access)
    expire_on = DateTimeUtils.get_expiration_date(expire_in).strftime('%Y-%m-%d')

    logging.getLogger().info('Creating public link. %s', {'path': str(path.relativepath), 'access': access, 'expire_on': expire_on})

    param = CreateShareParam.instance(path=path.fullpath(), access=access, expire_on=expire_on)

    response = ctera_host.execute('', 'createShare', param)

    return response.publicLink



import logging

from ..common import Object
from ..lib.session_base import SessionBase, SessionUser, SessionStatus, SessionHostType


class SessionType:
    Local = 'local'
    Remote = 'remote'


class SessionConnection(Object):
    def __init__(self, session_type, remote_from=None):
        self.type = session_type
        if session_type == SessionType.Remote:
            self.remote_from = remote_from
            self.remote_access = False
        else:
            self.remote_from = None
            self.remote_access = None


class LocalUser(SessionUser):
    """Local Session Object"""


class RemoteUser(SessionUser):
    """Remote Session Object"""

    def __init__(self, name, tenant):
        super().__init__(name)
        self.tenant = tenant


class Session(SessionBase):

    def __init__(self, host):
        super().__init__(host, SessionHostType.Edge)
        self.connection = SessionConnection(SessionType.Local)

    def _do_start_local_session(self, CTERA_Host):
        user = CTERA_Host.get('/currentuser').username
        self.set_version(CTERA_Host.get('/status/device/runningFirmware'))
        self.user = LocalUser(user)
        self.connection = SessionConnection(SessionType.Local)

    def start_remote_session(self, remote_session):
        self.user = RemoteUser(remote_session.user.name, remote_session.user.tenant)
        self.connection = SessionConnection(SessionType.Remote, remote_session.host)
        self.status = SessionStatus.Active

    def _do_terminate(self):
        if self.local():
            logging.getLogger().debug('Terminating local session. %s', {'host': self.host, 'user': self.user.name})
            self.connection = SessionConnection(SessionType.Local)
        elif self.remote():
            logging.getLogger().debug(
                'Terminating remote session. %s',
                {'host': self.host, 'tenant': self.user.tenant, 'user': self.user.name}
            )
            self.disable_remote_access()

    def local(self):
        return self.connection.type == SessionType.Local

    def remote(self):
        return self.connection.type == SessionType.Remote

    def enable_remote_access(self):
        self.connection.remote_access = True

    def disable_remote_access(self):
        self.connection.remote_access = False

    def remote_access(self):  # pylint: disable=method-hidden
        return self.connection.remote_access

    def remote_from(self):  # pylint: disable=method-hidden
        return self.connection.remote_from

    def tenant(self):
        return self.user.tenant if isinstance(self.user, RemoteUser) else None



__all__ = [
    'connection',
    'activation',
    'antivirus',
    'buckets',
    'directoryservice',
    'enum',
    'login',
    'query',
    'logs',
    'portals',
    'reports',
    'remote',
    'servers',
    'devices',
    'session',
    'admins',
    'users',
    'cloudfs',
    'files',
    'decorator',
    'messaging'
]



import logging

from .common import SrcDstParam, ActionResourcesParam


def delete(ctera_host, path):
    return delete_multi(ctera_host, *[path])


def delete_multi(ctera_host, *paths):
    delete_param = ActionResourcesParam.instance()

    for path in paths:
        logging.getLogger().info('Deleting item. %s', {'path': str(path.relativepath)})
        delete_param.add(SrcDstParam.instance(src=path.fullpath()))

    return ctera_host.execute('', 'deleteResources', delete_param)



from ..lib.session_base import SessionBase, SessionUser, SessionHostType
from .enum import Context, Role


class PortalUser(SessionUser):

    def __init__(self, name, tenant, role):
        super().__init__(name)
        self.tenant = tenant
        self.role = role


class Session(SessionBase):

    Administration = 'Administration'

    def __init__(self, host, context):
        super().__init__(host, SessionHostType.Portal)
        self.context = context
        self.local_auth = False

    def _do_start_local_session(self, CTERA_Host):
        tenant = CTERA_Host.get('/currentPortal') or Session.Administration
        self.set_version(CTERA_Host.get('/version'))
        if self.local_auth:
            self.user = PortalUser('$admin', tenant=tenant, role=Role.ReadWriteAdmin)
        else:
            current_session = CTERA_Host.get('/currentSession')
            self.user = PortalUser(current_session.username, tenant=tenant, role=current_session.role)

    def _do_terminate(self):
        pass

    def update_tenant(self, current_tenant):
        self.user.tenant = current_tenant or Session.Administration

    def is_global_admin(self):
        return self.context == Context.admin

    def is_local_auth(self):
        return self.local_auth

    def in_tenant_context(self):
        return self.tenant() != Session.Administration



import re
import os
import pathlib
from xml.dom import minidom
from xml.sax.saxutils import escape
from xml.parsers.expat import ExpatError


class Transcribe():
    COMMENT = '<!--TEMPLATE-->'
    SECTION_START = '<div style="border: 1px solid #EBECF0; margin-bottom: 10px;">'
    SECTION_END = '</div>'
    RECORDED_HEADERS = ['content-type', 'cookie']

    def __init__(self):
        self._filename = './CTERA_' + str(os.getpid()) + '.html'

    def transcribe(self, request, response=None):
        self._append_content(self._get_content(request, response))

    @classmethod
    def _get_content(cls, request, response):
        return ''.join(
            [
                Transcribe.SECTION_START,
                cls._get_request_content(request),
                cls._get_response_content(response) if response else '',
                Transcribe.SECTION_END
            ]
        )

    @classmethod
    def _get_request_content(cls, request):
        request_content_array = [
            '<p class="h4">' + request.method + ' ' + cls._parse_url(request.url) + '</p>'
        ]

        # Request headers:
        headers = cls._get_headers_array(request.headers)
        if len(headers) > 0:
            request_content_array.append("<br/><p class=\"h6\">Request headers:</p>")
            for key, value in headers:
                request_content_array.append('<p>' + key + ': ' + value + '</p>')

        # Request body:
        if request.body:
            request_content_array.append("<br/><p class=\"h6\">Request body:</p>")
            request_content_array.append('<pre>' + cls._prettify(request.body.decode('utf-8')) + '</pre>')

        return ''.join(request_content_array)

    @staticmethod
    def _parse_url(url):
        uri = url[re.search("https?://[^/]*", url).end():]
        return uri[:-1] if uri.endswith('?') else uri

    @classmethod
    def _get_headers_array(cls, request_headers):
        return [
            (key, value) for key, value in request_headers.items() if key.lower() in cls.RECORDED_HEADERS
        ]

    @classmethod
    def _get_response_content(cls, response):
        response_content_array = []
        reply = response.read()
        if reply:
            response_content_array.append("<br/><p class=\"h6\">Response body:</p>")
            response_content_array.append('<pre>' + cls._prettify(reply) + '</pre>')
        return ''.join(response_content_array)

    def _append_content(self, content):
        header, footer = self._get_header_footer()
        with open(self._filename, 'w', encoding='utf-8') as f:
            f.write(header + content + Transcribe.COMMENT + footer)

    def _get_header_footer(self):
        filename = self._filename if os.path.exists(self._filename) else \
            os.path.join(pathlib.Path(__file__).parent.absolute(), 'apidoc.template')
        with open(filename, 'r', encoding='utf-8') as f:
            content = f.read()
        array = content.split(Transcribe.COMMENT)
        return array[0], array[1]

    @staticmethod
    def _prettify(data):
        try:
            data = minidom.parseString(data).toprettyxml(indent="   ")
            data = ''.join(data.split('\n', 1)[1:])
            data = escape(data)
            return data
        except ExpatError:
            return data


def transcribe(request, response=None):
    Transcribe().transcribe(request, response)



import logging


def move(CTERAHost, src, dest, overwrite):
    source = src.fullpath()
    destination = dest.joinpath(src.name()).fullpath()
    logging.getLogger().info('Moving. %s', {'from': source, 'to': dest.fullpath()})
    CTERAHost.move(CTERAHost.make_local_files_dir(source), CTERAHost.make_local_files_dir(destination), overwrite, use_file_url=True)
    logging.getLogger().info('Moved. %s', {'from': source, 'to': dest.fullpath(), 'fullpath': destination})



import logging


def copy(CTERAHost, src, dest, overwrite):
    source = src.fullpath()
    destination = dest.joinpath(src.name()).fullpath()
    logging.getLogger().info('Copying. %s', {'from': source, 'to': dest.fullpath()})
    CTERAHost.copy(CTERAHost.make_local_files_dir(source), CTERAHost.make_local_files_dir(destination), overwrite, use_file_url=True)
    logging.getLogger().info('Copied. %s', {'from': source, 'to': dest.fullpath(), 'fullpath': destination})



import logging


def test(CTERAHost):
    test_network(CTERAHost)
    test_application(CTERAHost)


def test_network(CTERAHost):
    CTERAHost.test_conn()


def test_application(CTERAHost):
    ref = '/nosession/logininfo'

    logging.getLogger().debug('Trying to obtain login info. %s', {'ref': ref})

    response = CTERAHost.get(ref)

    logging.getLogger().debug('Successfully obtained login info. %s', {'ref': ref, 'hostname': response.hostname})

    return response



import logging


class Iterator:
    """ Objects Iterator """

    def __init__(self, function, param):
        self._function = function
        self._param = param
        self._hasMore = True
        self._objects = []

    def __iter__(self):
        return self

    def __next__(self):
        if not self._objects:
            if self._hasMore:
                self._hasMore, self._objects = self._function(self._param)
                if not self._hasMore and not self._objects:
                    self._terminate()
                self._param.increment()
            else:
                self._terminate()
        return self._objects.pop(0)

    @staticmethod
    def _terminate():
        logging.getLogger().debug('No more objects to return. Stopping iteration.')
        raise StopIteration



import logging
import time

from ..exception import CTERAException


class ErrorStatus(CTERAException):

    def __init__(self, status):
        super().__init__()
        self.status = status


class StatusTracker:

    def __init__(self, CTERAHost, ref, success, progress, transient, failure, retries, seconds):
        self.CTERAHost = CTERAHost
        self.ref = ref
        self.success = success
        self.progress = progress
        self.transient = transient
        self.failure = failure
        self.retries = retries
        self.seconds = seconds
        self.attempt = 0
        self.status = None

    def track(self):
        running = True
        while running:
            logging.getLogger().debug('Retrieving status. %s', {'ref': self.ref, 'attempt': (self.attempt + 1)})
            self.status = self.CTERAHost.get(self.ref)
            logging.getLogger().debug('Current status. %s', {'ref': self.ref, 'status': self.status})
            self.increment()
            running = self.running()
        return self.resolve()

    def resolve(self):
        if self.successful():
            logging.getLogger().debug('Success. %s', {'ref': self.ref, 'status': self.status})
            return self.status

        if self.failed():
            logging.getLogger().debug('Failure. %s', {'ref': self.ref, 'status': self.status})
            raise ErrorStatus(self.status)

        logging.getLogger().debug('Unknown status. %s', {'ref': self.ref, 'status': self.status})
        raise CTERAException('Unknown status', None, status=self.status)

    def successful(self):
        return self.status in self.success

    def running(self):
        if self.status in self.progress:
            logging.getLogger().debug('In progress. %s', {'ref': self.ref, 'status': self.status})
            return True

        if self.status in self.transient:
            logging.getLogger().debug('Transient state. %s', {'ref': self.ref, 'status': self.status})
            return True

        logging.getLogger().debug('End state. %s', {'ref': self.ref, 'status': self.status})
        return False

    def failed(self):
        return self.status in self.failure

    def increment(self):
        self.attempt = self.attempt + 1
        if self.attempt >= self.retries:
            logging.getLogger().error('Status did not meet success criteria. %s', {'ref': self.ref, 'status': self.status})
            raise CTERAException('Timed out. Status did not meet success criteria', None, ref=self.ref, status=self.status)

        logging.getLogger().debug('Sleep. %s', {'seconds': self.seconds})
        time.sleep(self.seconds)


def track(CTERAHost, ref, success, progress, transient, failure, retries=300, seconds=1):
    tracker = StatusTracker(CTERAHost, ref, success, progress, transient, failure, retries, seconds)
    return tracker.track()



import logging

from ..lib.task_manager_base import TaskError
from .licenses import Licenses
from ..lib import ask, track
from ..common import Object
from ..exception import CTERAException, InputError, CTERAConnectionError
from .. import config
from . import enum
from .base_command import BaseCommand
from .types import TCPService


class Services(BaseCommand):
    """ Gateway Cloud Services configuration APIs """
    _UNTRUSTED_CERTIFICATE_ERRORS = [
        'UntrustedCert',
        'UntrustedCA',
        'UntrustedCertExpired',
        'UntrustedCertSelfSigned',
        'UntrustedCertDNS'
    ]

    def __init__(self, gateway):
        super().__init__(gateway)
        self._trust_cert = {}

    def get_status(self):
        """
        Retrieve the cloud services connection status
        """
        status = self._gateway.get('/status/services')
        connection = Object()
        connection.connected = status.CTERAPortal.connectionState == enum.ServicesConnectionState.Connected
        if connection.connected:
            connection.ipaddr = status.CTERAPortal.connectedAddress
            connection.user = status.userDisplayName
            connection.server_version = status.portalVersion
            connection.server_address = status.CTERAPortal.serverList[0].name
            connection.last_connected_at = status.CTERAPortal.establishedTime
        return connection

    def connected(self):
        """
        Check if the Edge Filer is connected to CTERA Portal
        """
        return self._gateway.get('/status/services/CTERAPortal/connectionState') == enum.ServicesConnectionState.Connected

    def connect(self, server, user, password, ctera_license=enum.License.EV16):
        """
        Connect to a Portal.\n
        The connect method will first validate the `license` argument,
         ensure the Gateway can establish a TCP connection over port 995 to `server` using :py:func:`Gateway.tcp_connect()` and
         verify the Portal does not require device activation via code

        :param str server: Address of the Portal
        :param str user: User for the Portal connection
        :param str password: Password for the Portal connection
        :param cterasdk.edge.enum.License,optional ctera_license: CTERA License, defaults to cterasdk.edge.enum.License.EV16
        """
        Services._validate_license(ctera_license)
        self._check_cttp_traffic(address=server)
        self._check_connection(server)

        param = Object()
        param.server = server
        param.user = user
        param.password = password
        param.trustCertificate = self._trust_cert.get(server, False)
        self._connect_to_services(param, ctera_license)

    def activate(self, server, user, code, ctera_license=enum.License.EV16):
        """
        Activate the gateway using an activation code

        :param str server: Address of the Portal
        :param str user: User for  the Portal connection
        :param str code: Activation code for the Portal connection
        :param cterasdk.edge.enum.License,optional ctera_license: CTERA License, defaults to cterasdk.edge.enum.License.EV16
        """
        Services._validate_license(ctera_license)
        self._check_cttp_traffic(address=server)

        param = Object()
        param.server = server
        param.user = user
        param.activationCode = code
        param.trustCertificate = self._trust_cert.get(server, False)
        self._connect_to_services(param, ctera_license)

    def reconnect(self):
        """ Reconnect to the Portal """
        self._gateway.execute("/status/services", "reconnect", None)

    def disconnect(self):
        """ Disconnect from the Portal """
        self._gateway.put('/config/services', None)

    def sso_enabled(self):
        """
        Is SSO connection enabled

        :return bool: True if SSO connection is enabled, else False
        """
        return self._gateway.get('/config/gui/adminRemoteAccessSSO')

    def enable_sso(self):
        """ Enable SSO connection """
        self._set_sso(True)

    def disable_sso(self):
        """ Disable SSO connection """
        self._set_sso(False)

    def _set_sso(self, sso_state):
        logging.getLogger().info('%s single sign-on from CTERA Portal.', ('Enabling' if sso_state else 'Disabling'))
        self._gateway.put('/config/gui/adminRemoteAccessSSO', sso_state)
        logging.getLogger().info('Single sign-on %s.', ('enabled' if sso_state else 'disabled'))

    def _connect_to_services(self, param, ctera_license):
        task = self._attach(param)
        try:
            self._gateway.tasks.wait(task)
            track(self._gateway, '/status/services/CTERAPortal/connectionState', [enum.ServicesConnectionState.Connected],
                  [enum.ServicesConnectionState.ResolvingServers, enum.ServicesConnectionState.Connecting,
                   enum.ServicesConnectionState.Attaching, enum.ServicesConnectionState.Authenticating],
                  [], [enum.ServicesConnectionState.Disconnected], 20, 1)
            logging.getLogger().info("Connected to Portal.")
        except TaskError as error:
            description = error.task.description
            logging.getLogger().error("Connection failed. Reason: %s", description)
            raise CTERAException("Connection failed", None, reason=description)
        self._gateway.licenses.apply(ctera_license)

    @staticmethod
    def _validate_license(ctera_license):
        try:
            Licenses.infer(ctera_license)
        except InputError as error:
            logging.getLogger().error('Connection failed. Invalid license type. %s', {'license': ctera_license})
            raise error

    def _check_cttp_traffic(self, address, port=995):
        tcp_connect_result = self._gateway.network.tcp_connect(TCPService(address, port))
        if not tcp_connect_result.is_open:
            logging.getLogger().error("Unable to establish connection over port %s", str(tcp_connect_result.port))
            raise CTERAConnectionError('Unable to establish connection', None, host=tcp_connect_result.host,
                                       port=tcp_connect_result.port, protocol='CTTP')

    def _check_connection(self, server):
        param = Object()
        param.server = server
        param.trustCertificate = self._trust_cert.get(server, False)
        obj = self._gateway.execute('/status/services', 'isWebSsoEnabled', param)
        if not Services._check_web_sso(obj):
            self._handle_untrusted_cert(server, obj)

    @staticmethod
    def _check_web_sso(obj):
        try:
            if obj.result.hasWebSSO:
                message = "Connection failed. You must activate this Gateway using an activation code."
                logging.getLogger().error(message)
                raise CTERAException(message)

            if not obj.result.hasWebSSO and obj.rc == "connectOK":
                return True
            raise CTERAException("Connection failed", None, reason=obj.rc)
        except AttributeError:
            pass

        return False

    def _handle_untrusted_cert(self, server, obj):
        try:
            if obj.rc in Services._UNTRUSTED_CERTIFICATE_ERRORS:
                proceed = False
                if config.connect['ssl'] == 'Consent':
                    logging.getLogger().warning(msg=obj.msg)
                    proceed = ask("Proceed connecting '" + self._gateway.host() + "' to " + server + '?')
                if config.connect['ssl'] == 'Trust' or proceed:
                    self._trust_cert[server] = True
                    return self._check_connection(server)
        except AttributeError:
            pass
        return False

    def _attach(self, param):
        logging.getLogger().info("Connecting to Portal. %s", {'server': param.server, 'user': param.user})
        obj = self._gateway.execute("/status/services", "attachAndSave", param)
        return obj.id



import logging

from ..common import union
from ..exception import CTERAException, ObjectNotFoundException
from .base_command import BaseCommand
from . import query


class Buckets(BaseCommand):
    """
    Portal Storage Node APIs
    """

    default = ['name']

    def _get_entire_object(self, name):
        try:
            return self._portal.get(f'/locations/{name}')
        except CTERAException as error:
            raise CTERAException('Failed to get bucket', error)

    def _get_tenant_base_object_ref(self, name):
        return self._portal.portals.get(name, include=['baseObjectRef']).baseObjectRef

    def get(self, name, include=None):
        """
        Get a Bucket

        :param str name: Name of the bucket
        :param list[str] include: List of fields to retrieve, defaults to ``['name']``
        """
        include = union(include or [], Buckets.default)
        include = ['/' + attr for attr in include]
        bucket = self._portal.get_multi('/locations/' + name, include)
        if bucket.name is None:
            raise ObjectNotFoundException('Could not find bucket', f'/locations/{name}', name=name)
        return bucket

    def add(self, name, bucket, read_only=False, dedicated_to=None):
        """
        Add a Bucket

        :param str name: Name of the bucket
        :param cterasdk.core.types.Bucket bucket: Storage bucket to add
        :param bool,optional read_only: Set bucket to read-delete only, defaults to False
        :param str,optional dedicated_to: Name of a tenant, defaults to ``None``
        """
        param = bucket.to_server_object()
        param.name = name
        param.readOnly = read_only
        param.dedicated = bool(dedicated_to)
        param.dedicatedPortal = self._get_tenant_base_object_ref(dedicated_to) if dedicated_to else None

        logging.getLogger().info('Adding bucket. %s', {'name': name, 'bucket': bucket.bucket, 'type': bucket.__class__.__name__})
        response = self._portal.add('/locations', param)
        logging.getLogger().info('Bucket added. %s', {'name': name, 'bucket': bucket.bucket, 'type': bucket.__class__.__name__})
        return response

    def modify(self, current_name, new_name=None, read_only=None, dedicated_to=None):
        """
        Modify a Bucket

        :param str current_name: The current bucket name
        :param str,optional new_name: New name
        :param bool,optional read_only: Set bucket to read-delete only
        :param bool,optional dedicated: Dedicate bucket to a tenant
        :param str,optional dedicated_to: Tenant name
        """
        param = self._get_entire_object(current_name)
        if new_name:
            param.name = new_name
        if read_only is not None:
            param.readOnly = read_only
        if dedicated_to is not None:
            if isinstance(dedicated_to, bool):
                if not dedicated_to:
                    param.dedicated = False
                    param.dedicatedPortal = None
                else:
                    raise ValueError("'dedicated_to' must be either False or a 'str'")
            elif isinstance(dedicated_to, str):
                param.dedicated = True
                param.dedicatedPortal = self._get_tenant_base_object_ref(dedicated_to) if dedicated_to else None
        logging.getLogger().info("Modifying bucket. %s", {'name': current_name})
        response = self._portal.put(f'/locations/{current_name}', param)
        logging.getLogger().info("Bucket modified. %s", {'name': current_name})
        return response

    def list_buckets(self, include=None):
        """
        List Buckets.\n
        To retrieve buckets, you must first browse the Global Administration Portal, using: `GlobalAdmin.portals.browse_global_admin()`

        :param list[str],optional include: List of fields to retrieve, defaults to ``['name']``
        """
        include = union(include or [], Buckets.default)
        param = query.QueryParamBuilder().include(include).build()
        return query.iterator(self._portal, '/locations', param)

    def delete(self, name):
        """
        Delete a Bucket

        :param str name: Name of the bucket
        """
        logging.getLogger().info('Deleting bucket. %s', {'name': name})
        response = self._portal.delete(f'/locations/{name}')
        logging.getLogger().info('Bucket deleted. %s', {'name': name})
        return response

    def read_write(self, name):
        """
        Set bucket to Read Write

        :param str name: Name of the bucket
        """
        logging.getLogger().info('Setting bucket to read-write. %s', {'name': name})
        return self._read_only(name, False)

    def read_only(self, name):
        """
        Set bucket to Read Only

        :param str name: Name of the bucket
        """
        logging.getLogger().info('Setting bucket to read-delete only. %s', {'name': name})
        return self._read_only(name, True)

    def _read_only(self, name, enabled):
        return self._portal.put(f'/locations/{name}/readOnly', enabled)



import logging

from .base_command import BaseCommand


class CLI(BaseCommand):
    """ Gateway CLI APIs """

    def run_command(self, cli_command):
        """
        Run a CLI command

        :param str cli_command: The CLI command to run on the gateway
        :return str: The response of the gateway
        """
        logging.getLogger().info("Executing CLI command. %s", {'cli_command': cli_command})

        response = self._gateway.execute('/config/device', 'debugCmd', cli_command)

        logging.getLogger().info("CLI command executed. %s", {'cli_command': cli_command})

        return response



class Context:
    """
    Portal connection context

    :ivar str admin: Global admin context
    :ivar str ServicesPortal: Services Portal context
    """
    admin = 'admin'
    ServicesPortal = 'ServicesPortal'


class LogTopic:
    """
    Portal Log Topic

    :ivar str System: System log topic
    :ivar str CloudBackup: Cloud Backup log topic
    :ivar str CloudSync: Cloud Sync log topic
    :ivar str Access: Access log topic
    :ivar str Audit: Audit log topic
    """
    System = 'system'
    CloudBackup = 'backup'
    CloudSync = 'cloudsync'
    Access = 'access'
    Audit = 'audit'


class OriginType:
    """
    Log Origin Type

    :ivar str Portal: Portal originated logs
    :ivar str Device: Device originated logs
    """
    Portal = 'Portal'
    Device = 'Device'


class DeviceType:
    """
    Device type

    :ivar str CloudPlug: Cloud Plug device
    :ivar str C200: C200 device
    :ivar str C400: C400 device
    :ivar str C800: C800 device
    :ivar str C800P: C800P device
    :ivar str vGateway: vGateway device
    :ivar str ServerAgent: Server Agent device
    :ivar str WorkstationAgent: Workstation Agent Agent device
    :ivar list[str] Gateways: List of all the Gateway DeviceTypes
    :ivar list[str] Agents: List of all the Agents DeviceTypes
    """
    CloudPlug = "CloudPlug"
    C200 = "C200"
    C400 = "C400"
    C800 = "C800"
    C800P = "C800P"
    vGateway = "vGateway"
    ServerAgent = "Server Agent"
    WorkstationAgent = "Workstation Agent"
    Gateways = [CloudPlug, C200, C400, C800, C800P, vGateway]
    Agents = [ServerAgent, WorkstationAgent]


class Role:
    """
    Portal User Role

    :ivar str Disabled: Disabled user role
    :ivar str EndUser: EndUser user role
    :ivar str ReadWriteAdmin: ReadWriteAdmin user role
    :ivar str ReadOnlyAdmin: ReadOnlyAdmin user role
    :ivar str Support: Support user role
    """
    Disabled = "Disabled"
    EndUser = "EndUser"
    ReadWriteAdmin = "ReadWriteAdmin"
    ReadOnlyAdmin = "ReadOnlyAdmin"
    Support = "Support"


class RoleResolver:
    """
    Role Settings Resolver

    :ivar str ReadWriteAdmin: ReadWriteAdmin user role
    :ivar str ReadOnlyAdmin: ReadOnlyAdmin user role
    :ivar str Support: Support user role
    """
    ReadWriteAdmin = "readWriteAdminSettings"
    ReadOnlyAdmin = "readOnlyAdminSettings"
    Support = "supportAdminSettings"
    ComplianceOfficer = 'complianceOfficerSettings'


class Severity:
    """
    Portal Log Severity

    :ivar str EMERGENCY: Emergency log severity
    :ivar str ALERT: Alert log severity
    :ivar str CRITICAL: Critical log severity
    :ivar str ERROR: Error log severity
    :ivar str WARNING: Warning log severity
    :ivar str NOTICE: Notice log severity
    :ivar str INFO: Info log severity
    :ivar str DEBUG: Debug log severity
    """
    EMERGENCY = "emergency"
    ALERT = "alert"
    CRITICAL = "critical"
    ERROR = "error"
    WARNING = "warning"
    NOTICE = "notice"
    INFO = "info"
    DEBUG = "debug"


class PortalAccountType:
    """
    Portal Account Type

    :ivar str User: User account type
    :ivar str Group: Group account type
    """
    User = "user"
    Group = "group"


class SearchType:
    """
    Search Type

    :ivar str User: User search type
    :ivar str Group: Group search type
    """
    Users = "users"
    Groups = "groups"


class PolicyType:
    """
    Zone Policy Type

    :ivar str ALL: All folders
    :ivar str SELECT: Selected Folders
    :ivar str NONE: No Folders
    """
    ALL = 'allFolders'
    SELECT = 'selectedFolders'
    NONE = 'noFolders'


class ProtectionLevel:
    """
    External Share Protection Level

    :ivar str publicLink: No authentication
    :ivar str email: 2FA via email
    """
    Public = 'publicLink'
    Email = 'email'


class FileAccessMode:
    """
    Share Access Mode

    :ivar str RW: Read Write
    :ivar str RO: Read Only
    :ivar str PO: Preview Only
    :ivar str NA: None
    """
    RW = "ReadWrite"
    RO = "ReadOnly"
    PO = "PreviewOnly"
    NA = "None"


class CollaboratorType:
    """
    Collaborator Type

    :ivar str LU: Local User
    :ivar str DU: Domain User
    :ivar str LG: Local Group
    :ivar str DG: Domain Group
    :ivar str EXT: External
    """
    LU = "localUser"
    LG = "localGroup"
    DU = "adUser"
    DG = "adGroup"
    EXT = "external"


class PortalType:
    """
    Portal Type

    :ivar str Team: Team Portal
    :ivar str Reseller: Reseller Portal
    """
    Team = 'team'
    Reseller = 'reseller'


class ServerMode:
    """
    Portal Server Mode

    :ivar str Master: Master
    :ivar str Slave: Slave
    """
    Master = 'master'
    Slave = 'slave'


class SetupWizardStatus:
    """
    Portal Setup Wizard Status

    :ivar str NA: Not Relevant
    :ivar str Running: In Progress
    :ivar str Failed: Failed
    :ivar str Completed: Completed
    """
    NA = "notRelevant"
    Running = "inProgress"
    Failed = "failed"
    Completed = "completed"


class SetupWizardStage:
    """
    Portal Setup Wizard Stage

    :ivar str Server: Initializing Server
    :ivar str Portal: Initializing Portal
    :ivar str Replication: Setting Database Replication
    :ivar str Restart: Restarting Server
    :ivar str Finish: Finished
    """
    Server = "initServer"
    Portal = "initPortal"
    Replication = "setReplication"
    Restart = "restartingServer"
    Finish = "finish"


class SlaveAuthenticaionMethod:
    """
    Secondary Portal server authentication mode

    :ivar str Password: Password
    :ivar str PrivateKey: Private Key
    """
    Password = 'Password'
    PrivateKey = 'Key'


class PlanRetention:
    """
    Portal plan retention policy

    :ivar str All: All versions
    :ivar str Hourly: Hourly versions
    :ivar str Daily: Daily versions
    :ivar str Weekly: Weekly versions
    :ivar str Monthly: Monthly versions
    :ivar str Quarterly: Quarterly versions
    :ivar str Yearly: Yearly versions
    :ivar str Deleted: Recycle bin
    """
    All = 'retainAll'
    Hourly = 'hourly'
    Daily = 'daily'
    Weekly = 'weekly'
    Monthly = 'monthly'
    Quarterly = 'quarterly'
    Yearly = 'yearly'
    Deleted = 'retainDeleted'


class PlanItem:
    """
    Portal plan item

    :ivar str EV4: EV4
    :ivar str EV8: EV8
    :ivar str EV16: EV16
    :ivar str EV32: EV32
    :ivar str EV64: EV64
    :ivar str EV128: EV128
    :ivar str WA: Workstation Agent
    :ivar str SA: Server Agent
    :ivar str Share: Cloud Drive
    :ivar str Connect: Cloud Drive Connect
    """
    EV4 = 'EV4'
    EV8 = 'EV8'
    EV16 = 'EV16'
    EV32 = 'EV32'
    EV64 = 'EV64'
    EV128 = 'EV128'
    WA = 'WA'
    SA = 'SA'
    Share = 'Share'
    Connect = 'Connect'
    Storage = 'Storage'


class ListFilter:
    """
    Cloud Drive Folder List Filter

    :ivar str All: All
    :ivar str Deleted: Deleted
    :ivar str NonDeleted: NonDeleted
    """
    All = 'All'
    Deleted = 'Deleted'
    NonDeleted = 'NonDeleted'


class PlanCriteria:
    """
    Subscription Plan Auto Assignment Rule Builder Criterias

    :ivar str Username: Username
    :ivar str Groups: User groups
    :ivar str Role: User role
    :ivar str First: User first name
    :ivar str Last: User last name
    :ivar str Company: User company
    :ivar str BillingId: User billing id
    :ivar str Comment: User comment
    """
    Username = 'username'
    Groups = 'userGroups'
    Role = 'role'
    First = 'firstName'
    Last = 'lastName'
    Company = 'company'
    BillingId = 'billingId'
    Comment = 'comment'


class AntivirusType:
    """
    Antivirus Type

    :ivar str McAfeeWG: McAfee Web Gateway
    :ivar str Symantec: Symantec Protection Engine
    :ivar str ESET: ESET Gateway Security
    :ivar str Sophos: Sophos AV
    :ivar str McAfeeVSES: McAfee VirusScan Enterprise for Storage
    :ivar str TrendMicro: Trend Micro InterScan
    """
    McAfeeWG = 'McAfee'
    Symantec = 'Symantec'
    ESET = 'Eset'
    Sophos = 'Sophos'
    McAfeeVSES = 'McAfeeVSES'
    TrendMicro = 'TrendMicro'


class ICAPServices:
    """
    ICAP Services

    :ivar str Antivirus: Antivirus
    :ivar str DLP: Data Loss Prevention
    """
    Antivirus = "Antivirus"
    DLP = "DLP"


class LocationType:
    """
    Location Type

    :ivar str Azure: Azure Blob Storage
    :ivar str S3: Amazon Web Services S3
    :ivar str S3Compatible: S3 Compatible
    :ivar str NetAppStorageGRID: NetApp StorageGRID WebScale (S3)
    """
    Azure = 'AzureLocation'
    S3 = 'S3Location'
    S3Compatible = 'S3Compatible'
    NetAppStorageGRID = 'NetAppLocation'


class BucketType:
    """
    Bucket Type

    :ivar str Azure: Azure
    :ivar str Scality: Scality
    :ivar str AWS: Amazon Web Services S3
    :ivar str ICOS: IBM Cloud Object Storage
    :ivar str GenericS3: Generic S3
    :ivar str Nutanix: Nutanix S3
    :ivar str Wasabi: Wasabi S3
    :ivar str Google: Google S3
    :ivar str NetAppStorageGRID: NetApp StorageGRID WebScale (S3)
    """
    Azure = 'Azure'
    Scality = 'ScalityS3'
    AWS = 'S3'
    ICOS = 'CleverSafeS3'
    GenericS3 = 'GenericS3'
    Nutanix = 'Nutanix'
    Wasabi = 'WasabiS3'
    Google = 'GoogleS3'
    NetAppStorageGRID = 'NTAP'


class EnvironmentVariables:
    """
    Environment Variables.\n
    Some environment variables are applicable across platforms (i.e. Windows, Linux), while others are limited to a designated platform

    :ivar str ALLUSERSPROFILE: All users profile
    :ivar str WINDIR: Windows directory
    :ivar str TEMP: Temp directory
    :ivar str SYSTEMDRIVE: System drive
    :ivar str PROGRAMFILES: Program files
    :ivar str APPDATA: Application data
    :ivar str USERPROFILE: Current user profile
    :ivar str PRIMARYUSER: Primary user
    :ivar str USERS: Users directory (CTERA Edge Filer)
    :ivar str AGENTS: Agents directory (CTERA Edge Filer)
    :ivar str SYNCS: Syncs directory (CTERA Edge Filer)
    :ivar str PROJECTS: Projects directory (CTERA Edge Filer)
    """
    ALLUSERSPROFILE = '$ALLUSERSPROFILE'
    WINDIR = '$WINDIR'
    TEMP = '$TEMP'
    SYSTEMDRIVE = '$SYSTEMDRIVE'
    PROGRAMFILES = '$PROGRAMFILES'
    APPDATA = '$APPDATA'
    USERPROFILE = '$USERPROFILE'
    USERS = '$USERS'
    AGENTS = '$AGENTS'
    SYNCS = '$SYNCS'
    PROJECTS = '$PROJECTS'
    PRIMARYUSER = '$PRIMARYUSER'


class Platform:
    """
    CTERA Edge Platform Type.\n

    :ivar str C200_Orion: All users profile
    :ivar str C200_ARM: Windows directory
    :ivar str C200_Kirkwood: Temp directory
    :ivar str C400_C800: System drive
    :ivar str Edge_6: CTERA 6.0 Edge Filer
    :ivar str Edge_7: CTERA 7.0 Edge Filer
    :ivar str Windows: Windows Agent (Drive App)
    :ivar str Linux: Linux Agent (Drive App)
    :ivar str OSX: Mac Agent (Drive App)
    """
    C200_Orion = 'Orion'
    C200_ARM = 'ARM'
    C200_Kirkwood = 'Kirkwood'
    C400_C800 = 'X86'
    Edge_6 = 'VBox'
    Edge_7 = 'Genesis'
    Linux = 'LinuxX86'
    Windows = 'WindowsX86'
    OSX = 'OSxX86'


class TemplateCriteria:
    """
    Configuration Template Auto Assignment Rule Builder Criterias

    :ivar str Type: Device type
    :ivar str OperatingSystem: Operating system
    :ivar str Version: Installed software version
    :ivar str Hostname: Hostname
    :ivar str Name: Device name
    :ivar str Owner: Device owner username
    :ivar str Plan: Plan name
    :ivar str Groups: Device owner local or domain groups
    """
    Type = 'DeviceType'
    OperatingSystem = 'OperatingSystem'
    Version = 'InstalledSoftwareVersion'
    Hostname = 'Hostname'
    Name = 'DeviceName'
    Owner = 'OwnerUsername'
    Plan = 'Plan'
    Groups = 'ownerGroups'


class IPProtocol:
    """
    IP Protocol

    :ivar str TCP: TCP Protocol
    :ivar str UDP: UDP Protocol
    """
    TCP = "TCP"
    UDP = "UDP"


class Mode:
    """
    Enum for operational mode

    :ivar str Enabled: Operational mode enabled
    :ivar str Disabled: Operational mode diabled
    """
    Enabled = "enabled"
    Disabled = "disabled"


class DirectoryServiceType:
    """
    Directory Service Type

    :ivar str Microsoft: Active Directory
    :ivar str LDAP: LDAP
    :ivar str Apple: Apple Open Directory
    """
    Microsoft = 'ActiveDirectory'
    LDAP = 'LDAP'
    Apple = 'AppleOpenDirectory'


class DirectoryServiceFetchMode:
    """
    Directory Service Fetch Mode

    :ivar str Eager: Eager
    :ivar str Lazy: Lazy
    """
    Eager = 'Eager'
    Lazy = 'Lazy'


class DirectorySearchEntityType:
    """
    Directory Search Entity Type

    :ivar str User: User
    :ivar str Group: Group
    """
    User = 'user'
    Group = 'group'


class DeduplicationMethodType:
    """
    Folder Group Deduplication Method Type

    :ivar str AverageBlockSize: AverageBlockSize
    :ivar str FixedBlockSize: FixedBlockSize
    """
    AverageBlockSize = 'AverageBlockSize'
    FixedBlockSize = 'FixedBlockSize'


class RetentionMode:
    """
    Write Once Read Many Retention Mode

    :ivar str Enterprise: Enterprise
    :ivar str Compliance: Compliance
    :ivar str Delete: Delete
    """
    Delete = 'None'
    Enterprise = 'Enterprise'
    Compliance = 'Compliance'


class Duration:
    """
    Duration

    :ivar str Minutes: Minutes
    :ivar str Hours: Hours
    :ivar str Days: Days
    :ivar str Months: Months
    :ivar str Years: Years
    """
    Minutes = 'Minutes'
    Hours = 'Hours'
    Days = 'Days'
    Months = 'Months'
    Years = 'Years'



import logging

from ...exception import CTERAException
from .path import CTERAPath


class ItemExists(CTERAException):
    pass


class Forbidden(CTERAException):
    pass


CreateDirectoryRC = {
    "File exists": ItemExists,
    "Creating a folder in this location is forbidden": Forbidden
}


def mkdir(ctera_host, path, recurse=False):
    if recurse:
        array = path.parts()
        for i in range(1, len(array) + 1):
            dirpath = CTERAPath('/'.join(array[:i]), path.basepath)
            try:
                _mkdir(ctera_host, dirpath)
            except (ItemExists, Forbidden):
                pass
    else:
        _mkdir(ctera_host, path)


def _mkdir(ctera_host, path):
    fullpath = path.fullpath()
    logging.getLogger().info('Creating directory. %s', {'path': fullpath})
    try:
        ctera_host.mkcol(ctera_host.make_local_files_dir(fullpath), use_file_url=True)
    except CTERAException as error:
        _process_error(error, fullpath)
    logging.getLogger().info('Directory created. %s', {'path': fullpath})


def _process_error(error, path):
    if hasattr(error.response.body, 'msg'):
        message = error.response.body.msg
        error = CreateDirectoryRC.get(message)
        if error is not None:
            error = error()
            error.message = message
            error.path = path
            raise error



import logging

from ..exception import CTERAException
from ..common import Object
from . import enum
from .base_command import BaseCommand


class SNMP(BaseCommand):
    """ Edge Filer SNMP Configuration APIs """

    def is_enabled(self):
        """
        Check if SNMP is enabled

        :return: True is SNMP is enabled, else False
        :rtype: bool
        """
        return self._gateway.get('/config/snmp/mode') == enum.Mode.Enabled

    def enable(self, port=161, community_str=None, username=None, auth_password=None, privacy_password=None):
        """
        Enable SNMP

        :param int,optional port: SNMP server port, defaults to 161
        :param str,optional community_str: SNMPv2c community string
        :param str,optional username: SNMPv3 username
        :param str,optional auth_password: SNMPv3 authentication password
        :param str,optional privacy_password: SNMPv3 privacy password
        """
        param = Object()
        param.mode = enum.Mode.Enabled
        param.port = port
        param.readCommunity = community_str
        if username is not None and auth_password is not None and privacy_password is not None:
            param.snmpV3 = Object()
            param.snmpV3.mode = enum.Mode.Enabled
            param.snmpV3.username = username
            param.snmpV3.authenticationPassword = auth_password
            param.snmpV3.privacyPassword = privacy_password

        logging.getLogger().info("Enabling SNMP.")
        self._gateway.put('/config/snmp', param)
        logging.getLogger().info("Enabled SNMP.")

    def disable(self):
        """ Disable SNMP """
        logging.getLogger().info("Disabling SNMP.")
        self._gateway.put('/config/snmp/mode', enum.Mode.Disabled)
        logging.getLogger().info("Disabled SNMP.")

    def get_configuration(self):
        return self._gateway.get('/config/snmp')

    def modify(self, port=None, community_str=None, username=None, auth_password=None, privacy_password=None):
        """
        Modify current SNMP configuration. Only configurations that are not `None` will be changed. SNMP must be enabled

        :param int,optional port: SNMP server port, defaults to 161
        :param str,optional community_str: SNMPv2c community string
        :param str,optional username: SNMPv3 username
        :param str,optional auth_password: SNMPv3 authentication password
        :param str,optional privacy_password: SNMPv3 privacy password
        """
        current_config = self.get_configuration()
        if current_config.mode == enum.Mode.Disabled:
            raise CTERAException("SNMP configuration cannot be modified when disabled")
        if port:
            current_config.port = port
        if community_str:
            current_config.readCommunity = community_str
        if username is not None and auth_password is not None and privacy_password is not None:
            current_config.snmpV3 = Object()
            current_config.snmpV3.mode = enum.Mode.Enabled
            current_config.snmpV3.username = username
            current_config.snmpV3.authenticationPassword = auth_password
            current_config.snmpV3.privacyPassword = privacy_password

        logging.getLogger().info("Updating SNMP configuration.")
        self._gateway.put('/config/snmp', current_config)
        logging.getLogger().info("SNMP configured.")



from .base_command import BaseCommand
from .enum import DeviceType
from . import remote, query
from ..common import union
from ..exception import CTERAException, ObjectNotFoundException


class Devices(BaseCommand):
    """ Portal Devices APIs """

    name_attr = 'name'
    type_attr = 'deviceType'
    default = ['name', 'portal', 'deviceType', 'version']

    def _create_device_resource_uri(self, device_name, tenant):
        session = self._portal.session()
        if not tenant:
            if not session.in_tenant_context():
                raise CTERAException('You must specify a tenant name or browse the tenant first.')
            tenant = self._portal.session().tenant()
        if session.is_local_auth():
            resource_uri = '/devices/' + device_name  # local auth: auto appends /portals/{tenant_name}
        else:
            resource_uri = f'/portals/{tenant}/devices/{device_name}'  # regular auth: support both tenant and Administration context
        return resource_uri

    def device(self, device_name, tenant=None, include=None):
        """
        Get a Device by its name

        :param str device_name: Name of the device to retrieve
        :param str,optional tenant: Tenant of the device, defaults to the tenant in the current session
        :param list[str],optional include: List of fields to retrieve, defaults to ['name', 'portal', 'deviceType']

        :return: Managed Device
        :rtype: ctera.object.Gateway.Gateway or ctera.object.Agent.Agent
        """
        include = union(include or [], Devices.default)
        include = ['/' + attr for attr in include]

        resource_uri = self._create_device_resource_uri(device_name, tenant)

        dev = self._portal.get_multi(resource_uri, include)
        if dev.name is None:
            raise ObjectNotFoundException('Device not found', resource_uri, tenant=tenant, name=device_name)

        return remote.remote_command(self._portal, dev)

    def filers(self, include=None, allPortals=False, deviceTypes=None):
        """
        Get Filers

        :param list[str],optional include: List of fields to retrieve, defaults to ['name', 'portal', 'deviceType']
        :param bool,optional allPortals: Search in all portals, defaults to False
        :param list[cterasdk.core.enum.DeviceType.Gateways] deviceTypes: Types of Filers, defaults to all Filer types

        :return: Iterator for all matching Filers
        :rtype: cterasdk.lib.iterator.Iterator[cterasdk.object.Gateway.Gateway]
        """
        if deviceTypes:
            deviceTypes = [deviceType for deviceType in deviceTypes if deviceType in DeviceType.Gateways]
        if not deviceTypes:
            deviceTypes = DeviceType.Gateways

        filters = [query.FilterBuilder(Devices.type_attr).eq(deviceType) for deviceType in deviceTypes]
        return self.devices(include, allPortals, filters)

    def agents(self, include=None, allPortals=False):
        """
        Get Agents

        :param list[str],optional include: List of fields to retrieve, defaults to ['name', 'portal', 'deviceType']
        :param bool,optional allPortals: Search in all portals, defaults to False

        :return: Iterator for all matching Agents
        :rtype: cterasdk.lib.iterator.Iterator[cterasdk.object.Agent.Agent]
        """
        filters = [query.FilterBuilder(Devices.type_attr).like('Agent')]
        return self.devices(include, allPortals, filters)

    def desktops(self, include=None, allPortals=False):
        """
        Get Desktops

        :param list[str],optional include: List of fields to retrieve, defaults to ['name', 'portal', 'deviceType']
        :param bool,optional allPortals: Search in all portals, defaults to False

        :return: Iterator for all matching Desktops
        :rtype: cterasdk.lib.iterator.Iterator
        """
        filters = [query.FilterBuilder(Devices.type_attr).eq(DeviceType.WorkstationAgent)]
        return self.devices(include, allPortals, filters)

    def servers(self, include=None, allPortals=False):
        """
        Get Servers

        :param list[str],optional include: List of fields to retrieve, defaults to ['name', 'portal', 'deviceType']
        :param bool,optional allPortals: Search in all portals, defaults to False

        :return: Iterator for all matching Servers
        :rtype: cterasdk.lib.iterator.Iterator
        """
        filters = [query.FilterBuilder(Devices.type_attr).eq(DeviceType.ServerAgent)]
        return self.devices(include, allPortals, filters)

    def by_name(self, names, include=None):
        """
        Get Devices by their names

        :param list[str],optional names: List of names of devices
        :param list[str],optional include: List of fields to retrieve, defaults to ['name', 'portal', 'deviceType']

        :return: Iterator for all matching Devices
        :rtype: cterasdk.lib.iterator.Iterator
        """
        filters = [query.FilterBuilder('name').eq(name) for name in names]
        return self.devices(include, False, filters)

    def devices(self, include=None, allPortals=False, filters=None, user=None):
        """
        Get Devices

        :param list[str],optional include: List of fields to retrieve, defaults to ['name', 'portal', 'deviceType']
        :param bool,optional allPortals: Search in all portals, defaults to False
        :param list[],optional filters: List of additional filters, defaults to None
        :param cterasdk.core.types.UserAccount user: User account of the device owner

        :return: Iterator for all matching Devices
        :rtype: cterasdk.lib.iterator.Iterator
        """
        include = union(include or [], Devices.default)
        builder = query.QueryParamBuilder().include(include).allPortals(allPortals)
        filters = filters or []
        for query_filter in filters:
            builder.addFilter(query_filter)
        if user:
            uid = self._portal.users.get(user, ['uid']).uid
            builder.ownedBy(uid)
        builder.orFilter((len(filters) > 1))
        param = builder.build()
        # Check if the _all attribute conflicts with the current tenant
        iterator = query.iterator(self._portal, '/devices', param)
        for dev in iterator:
            yield remote.remote_command(self._portal, dev)

    def get_comment(self, device_name, tenant=None):
        """
        Get Portal device comment

        :param str device: Device name
        :returns: Comment
        :rtype: str
        """
        return self._portal.get(f'{self._create_device_resource_uri(device_name, tenant)}/comment')

    def set_comment(self, device_name, comment, tenant=None):
        """
        Set a comment to a Portal device

        :param str device: Device name
        :param str comment: Comment
        """
        return self._portal.put(f'{self._create_device_resource_uri(device_name, tenant)}/comment', comment)



from .path import CTERAPath
from . import copy, move, mkdir, rm, file_access
from . import open as openfile


class FileBrowser:
    """ Gateway File Browser APIs """

    def __init__(self, Gateway):
        self._CTERAHost = Gateway
        self._file_access = file_access.FileAccess(Gateway)

    @staticmethod
    def ls(_path):
        return

    def openfile(self, path):
        """
        Obtain a file handle

        :param str path: The file path on the Edge Filer
        """
        return openfile.openfile(self._CTERAHost, self.mkpath(path))

    def download(self, path, destination=None):
        """
        Download a file

        :param str path: The file path on the Edge Filer
        :param str,optional destination:
         File destination, if it is a directory, the original filename will be kept, defaults to the default directory
        """
        return self._file_access.download(self.mkpath(path), destination=destination)

    def download_as_zip(self, cloud_directory, files, destination=None):
        """
        Download a list of files and/or directories from a cloud folder as a ZIP file

        .. warning:: The list of files is not validated. The ZIP file will include only the existing  files and directories

        :param str cloud_directory: Path to the cloud directory
        :param list[str] files: List of files and/or directories in the cloud folder to download
        :param str,optional destination:
         File destination, if it is a directory, the filename will be calculated, defaults to the default directory
        """
        return self._file_access.download_as_zip(self.mkpath(cloud_directory), files, destination=destination)

    def upload(self, file_path, server_path):
        """
        Upload a file

        :param str file_path: Path to the local file to upload
        :param str server_path: Path to the directory to upload the file to
        """
        self._file_access.upload(file_path, self.mkpath(server_path))

    def mkdir(self, path, recurse=False):
        """
        Create a new directory

        :param str path: The path of the new directory
        :param bool,optional recurse: Create subdirectories if missing, defaults to False
        """
        return mkdir.mkdir(self._CTERAHost, self.mkpath(path), recurse)

    def copy(self, src, dest, overwrite=False):
        """
        Copy a file or a folder

        :param str src: Source file or folder path
        :param str dest: Destination folder path
        :param bool,optional overwrite: Overwrite on conflict, defaults to False
        """
        return copy.copy(self._CTERAHost, self.mkpath(src), self.mkpath(dest), overwrite)

    def move(self, src, dest, overwrite=False):
        """
        Move a file or a folder

        :param str src: Source file or folder path
        :param str dest: Destination folder path
        :param bool,optional overwrite: Overwrite on conflict, defaults to False
        """
        return move.move(self._CTERAHost, self.mkpath(src), self.mkpath(dest), overwrite)

    def delete(self, path):
        """
        Delete a file

        :param str path: The file's path on the gateway
        """
        return rm.delete(self._CTERAHost, self.mkpath(path))

    @staticmethod
    def mkpath(path):
        return CTERAPath(path, '/')



import logging


def delete(ctera_host, path):
    fullpath = path.fullpath()
    logging.getLogger().info('Deleting item. %s', {'path': fullpath})

    response = ctera_host.rm(ctera_host.make_local_files_dir(fullpath))

    logging.getLogger().info('Item deleted. %s', {'path': fullpath})

    return response



import logging
from .path import CTERAPath

from ... import config
from ...exception import CTERAException
from ..base_command import BaseCommand
from . import ls, directory, rename, rm, recover, mv, cp, ln, collaboration, file_access


class FileBrowser(BaseCommand):

    """
    Portal File Browser APIs
    """

    def __init__(self, portal, base_path):
        super().__init__(portal)
        self._base_path = base_path
        self._file_access = file_access.FileAccess(portal)

    def ls(self, path, include_deleted=False):
        """
        Execute ls on the provided path

        :param str path: Path to list
        :param bool,optional include_deleted: Include deleted files, defaults to False
        """
        return ls.ls(self._portal, self.mkpath(path), include_deleted=include_deleted)

    def walk(self, path, include_deleted=False):
        """
        Perform walk on the provided path

        :param str path: Path to perform walk on
        :param bool,optional include_deleted: Include deleted files, defaults to False
        """
        paths = [self.mkpath(path)]

        while len(paths) > 0:
            path = paths.pop(0)
            items = ls.ls(self._portal, path, include_deleted=include_deleted)
            for item in items:
                if item.isFolder:
                    paths.append(self.mkpath(item))
                yield item

    def download(self, path, destination=None):
        """
        Download a file

        :param str path: Path of the file to download
        :param str,optional destination:
         File destination, if it is a directory, the original filename will be kept, defaults to the default directory
        """
        path = self.mkpath(path)
        return self._file_access.download(path, destination=destination)

    def download_as_zip(self, cloud_directory, files, destination=None):
        """
        Download a list of files and/or directories from a cloud folder as a ZIP file

        .. warning:: The list of files is not validated. The ZIP file will include only the existing  files and directories

        :param str cloud_directory: Path to the cloud directory
        :param list[str] files: List of files and/or directories in the cloud folder to download
        :param str,optional destination:
         File destination, if it is a directory, the original filename will be kept, defaults to the default directory
        """
        self._file_access.download_as_zip(self.mkpath(cloud_directory), files, destination=destination)

    def copy(self, src, dest):
        """
        Copy a file or directory

        :param str src: The source path of the file or directory
        :param str dst: The destination path of the file or directory
        """
        return cp.copy(self._portal, self.mkpath(src), self.mkpath(dest))

    def copy_multi(self, src, dest):
        return cp.copy_multi(self._portal, self.mkpath(src), self.mkpath(dest))

    def mklink(self, path, access='RO', expire_in=30):
        """
        Create a link to a file

        :param str path: The path of the file to create a link to
        :param str,optional access: Access policy of the link, defaults to 'RO'
        :param int,optional expire_in: Number of days until the link expires, defaults to 30
        """
        return ln.mklink(self._portal, self.mkpath(path), access, expire_in)

    def mkpath(self, array):
        if isinstance(array, list):
            return [CTERAPath(item, self._base_path) for item in array]
        return CTERAPath(array, self._base_path)


class CloudDrive(FileBrowser):

    """
    Cloud Drive File Browser APIs
    """

    def upload(self, file_path, server_path):
        """
        Upload a file

        :param str file_path: Path to the local file to upload
        :param str server_path: Path to the directory to upload the file to
        """
        self._file_access.upload(file_path, self.mkpath(server_path))

    def mkdir(self, path, recurse=False):
        """
        Create a new directory

        :param str path: Path of the directory to create
        :param bool,optional recurse: Whether to create the path recursivly, defaults to False
        """
        directory.mkdir(self._portal, self.mkpath(path), recurse)

    def rename(self, path, name):
        """
        Rename a file

        :param str path: Path of the file or directory to rename
        :param str name: The name to rename to
        """
        return rename.rename(self._portal, self.mkpath(path), name)

    def delete(self, path):
        """
        Delete a file

        :param str path: Path of the file or directory to delete
        """
        return rm.delete(self._portal, self.mkpath(path))

    def delete_multi(self, *args):
        """
        Delete multiple files and/or directories

        :param `*args`: Variable lengthed list of paths of files and/or directories to delete
        """
        return rm.delete_multi(self._portal, *self.mkpath(list(args)))

    def undelete(self, path):
        """
        Restore a previously deleted file or directory

        :param str path: Path of the file or directory to restore
        """
        return recover.undelete(self._portal, self.mkpath(path))

    def undelete_multi(self, *args):
        """
        Restore previously deleted multiple files and/or directories

        :param `*args`: Variable length list of paths of files and/or directories to restore
        """
        return recover.undelete_multi(self._portal, *self.mkpath(list(args)))

    def move(self, src, dest):
        """
        Move a file or directory

        :param str src: The source path of the file or directory
        :param str dst: The destination path of the file or directory
        """
        return mv.move(self._portal, self.mkpath(src), self.mkpath(dest))

    def move_multi(self, src, dest):
        return mv.move_multi(self._portal, self.mkpath(src), self.mkpath(dest))

    def get_share_info(self, path):
        """
        Get share settings and recipients
        """
        return collaboration.get_share_info(self._portal, self.mkpath(path))

    def share(self, path, recipients, as_project=True, allow_reshare=True, allow_sync=True):
        """
        Share a file or a folder

        :param str path: The path of the file or folder to share
        :param list[cterasdk.core.types.ShareRecipient] recipients: A list of share recipients
        :param bool,optional as_project: Share as a team project, defaults to True when the item is a cloud folder else False
        :param bool,optional allow_reshare: Allow recipients to re-share this item, defaults to True
        :param bool,optional allow_sync: Allow recipients to sync this item, defaults to True when the item is a cloud folder else False
        :return: A list of all recipients added to the collaboration share
        :rtype: list[cterasdk.core.types.ShareRecipient]
        """
        return collaboration.share(self._portal, self.mkpath(path), recipients, as_project, allow_reshare, allow_sync)

    def add_share_recipients(self, path, recipients):
        """
        Add share recipients

        :param str path: The path of the file or folder
        :param list[cterasdk.core.types.ShareRecipient] recipients: A list of share recipients
        :return: A list of all recipients added
        :rtype: list[cterasdk.core.types.ShareRecipient]
        """
        return collaboration.add_share_recipients(self._portal, self.mkpath(path), recipients)

    def remove_share_recipients(self, path, accounts):
        """
        Remove share recipients

        :param str path: The path of the file or folder
        :param list[cterasdk.core.types.PortalAccount] accounts: A list of portal user or group accounts
        :return: A list of all share recipients removed
        :rtype: list[cterasdk.core.types.PortalAccount]
        """
        return collaboration.remove_share_recipients(self._portal, self.mkpath(path), accounts)

    def unshare(self, path):
        """
        Unshare a file or a folder
        """
        return collaboration.unshare(self._portal, self.mkpath(path))


class Backups(FileBrowser):

    """
    Backups File Browser APIs
    """

    def device_config(self, device, destination=None):
        """
        Download a device configuration file

        :param str device: The device name
        :param str,optional destination:
         File destination, if it is a directory, the original filename will be kept, defaults to the default directory
        """
        try:
            destination = destination if destination is not None else f'{config.filesystem["dl"]}/{device}.xml'
            return self.download(f'{device}/Device Configuration/db.xml', destination)
        except CTERAException as error:
            logging.getLogger().error('Failed downloading configuration file. %s', {'device': device, 'error': error.response.reason})
            raise error



from ...common import Object


class FetchResourcesParam(Object):

    def __init__(self):
        self._classname = 'FetchResourcesParam'
        self.start = 0
        self.limit = 100

    def increment(self):
        self.start = self.start + self.limit


class FetchResourcesParamBuilder:

    def __init__(self):
        self.param = FetchResourcesParam()

    def root(self, root):
        self.param.root = root  # pylint: disable=attribute-defined-outside-init
        return self

    def depth(self, depth):
        self.param.depth = depth  # pylint: disable=attribute-defined-outside-init
        return self

    def include_deleted(self):
        self.param.includeDeleted = True  # pylint: disable=attribute-defined-outside-init

    def build(self):
        return self.param



from ..common import Object
from ..convert import tojsonstr


def query(CTERAHost, path, key, value):
    param = Object()
    param.key = key
    param.value = value
    return CTERAHost.db(path, 'query', param)


def show(CTERAHost, path, key, value):
    print(tojsonstr(query(CTERAHost, path, key, value), no_log=False))


class QueryParam(Object):

    def __init__(self):
        self.startFrom = 0
        self.countLimit = 50

    def include_classname(self):
        self._classname = self.__class__.__name__  # pylint: disable=attribute-defined-outside-init

    def increment(self):
        self.startFrom = self.startFrom + self.countLimit


class QueryParamBuilder:

    def __init__(self):
        self.param = QueryParam()

    def startFrom(self, startFrom):
        self.param.startFrom = startFrom
        return self

    def countLimit(self, countLimit):
        self.param.countLimit = countLimit
        return self

    def include(self, include):
        self.param.include = include  # pylint: disable=attribute-defined-outside-init
        return self

    def put(self, key, value):
        setattr(self.param, key, value)
        return self

    def build(self):
        return self.param



import functools
import logging

from ..exception import CTERAException


def authenticated(function):

    @functools.wraps(function)
    def check_authenticated_and_call(self, *args):
        session = self.session()

        if session.authenticated():
            return function(self, *args)
        if is_nosession(function, args[0]):
            return function(self, *args)
        logging.getLogger().error('Not logged in.')
        raise CTERAException('Not logged in')

    return check_authenticated_and_call


def is_nosession(function, path):
    return function.__name__ == 'get' and path.startswith('/nosession')



import logging
import time

from ..exception import HostUnreachable, ExhaustedException
from .base_command import BaseCommand


class Power(BaseCommand):
    """ Gateway Power APIs """

    def reboot(self, wait=False):
        """
        Reboot the Gateway

        :param bool,optional wait: Wait for reboot to complete, defaults to False
        """
        logging.getLogger().info("Rebooting device. %s", {'host': self._gateway.host()})
        self._gateway.execute("/status/device", "reboot", None)
        if wait:
            Boot(self._gateway).wait()

    def shutdown(self):
        """ Shutdown the Gateway """
        self._gateway.execute("/status/device", "poweroff", None)

    def reset(self, wait=False):
        """
        Reset the Gateway setting

        :param bool,optional wait: Wait for reset to complete, defaults to False
        """
        self._gateway.execute("/status/device", "reset2default", None)
        logging.getLogger().info("Resetting device to default settings. %s", {'host': self._gateway.host()})
        if wait:
            Boot(self._gateway).wait()


class Boot:

    def __init__(self, gateway, retries=60, seconds=5):
        self._gateway = gateway
        self._retries = retries
        self._seconds = seconds
        self._attempt = 0

    def wait(self):
        while True:
            try:
                self._increment()
                logging.getLogger().debug('Checking if device is up and running. %s', {'attempt': self._attempt})
                self._gateway.test()
                logging.getLogger().info("Device is back up and running.")
                break
            except (HostUnreachable, ExhaustedException) as e:
                logging.getLogger().debug('Exception. %s', {'exception': e.__class__.__name__, 'message': e.message})

    def _increment(self):
        self._attempt = self._attempt + 1
        if self._attempt >= self._retries:
            self._unreachable()
        logging.getLogger().debug('Sleep. %s', {'seconds': self._seconds})
        time.sleep(self._seconds)

    def _unreachable(self):
        scheme = self._gateway.scheme()
        host = self._gateway.host()
        port = self._gateway.port()

        logging.getLogger().error('Timed out. Could not reach host. %s', {'scheme': scheme, 'host': host, 'port': port})
        raise HostUnreachable(None, host, port, scheme)



import logging

from ..common import parse_base_object_ref
from ..exception import CTERAException


def api(Gateway):
    session = Gateway.session()
    if session.local():                             # direct access, via IP, hostname or FQDN
        return local(Gateway.baseurl())

    if session.remote():
        baseurl = Gateway._Portal.base_portal_url  # pylint: disable=protected-access
        device = Gateway.host()
        if session.remote_access():
            return remote_access(baseurl, device)   # remote: Gateway.remote_access()

        base_object_ref = parse_base_object_ref(Gateway.portal)
        return remote(baseurl, base_object_ref.name, device)  # remote, without a Gateway session

    logging.getLogger().error('Invalid connection type.')
    raise CTERAException('Invalid connection type', session)


def local(baseurl):
    return f'{baseurl}/admingui/api'


def remote(baseurl, tenant, device):
    return f'{baseurl}/devicecmdnew/{tenant}/{device}/'


def remote_access(baseurl, device):
    return f'{baseurl}/devices/{device}/admingui/api'


def files(Gateway):
    session = Gateway.session()
    if session.local():
        return Gateway.baseurl()

    if session.remote():
        if session.remote_access():
            baseurl = Gateway._Portal.base_portal_url  # pylint: disable=protected-access
            device = Gateway.host()
            return f'{baseurl}/devices/{device}'
    raise CTERAException('Invalid connection type', session)



import logging
import mimetypes
import shutil
import os
from pathlib import Path

from .. import config
from ..exception import RenameException, LocalDirectoryNotFound, LocalFileNotFound, LocalPathNotFound


class FileSystem:  # pylint: disable=unused-private-member

    __instance = None

    @staticmethod
    def instance():
        if FileSystem.__instance is None:
            FileSystem()
        return FileSystem.__instance

    def __init__(self):
        if FileSystem.__instance is not None:
            raise Exception("FileSystem is a singleton class.")
        FileSystem.__instance = self

    @staticmethod
    def expanduser(path):
        return os.path.expanduser(path)

    @staticmethod
    def exists(filepath):
        return os.path.exists(filepath)

    def rename(self, dirpath, src, dst):
        source = os.path.join(dirpath, src)
        if self.exists(source):
            destination = os.path.join(dirpath, dst)
            os.rename(source, destination)
            return (dirpath, dst)
        logging.getLogger().error('Could not rename temporary file. File not found. %s', {'path': dirpath, 'temp': src})
        raise RenameException(dirpath, src, dst)

    def validate_directory(self, dirpath):
        dirpath = os.path.expanduser(dirpath)
        if not self.exists(dirpath):
            raise LocalDirectoryNotFound(dirpath)

    def copyfile(self, src, dst):
        self.get_local_file_info(src)
        return shutil.copyfile(src, dst)

    def save(self, dirpath, filename, handle):
        dirpath = os.path.expanduser(dirpath)
        if not self.exists(dirpath):
            raise LocalDirectoryNotFound(dirpath)

        tempfile = filename + '.Chopin3'
        filepath = os.path.join(dirpath, tempfile)
        self.write(filepath, handle)
        origin = filename
        version = 0
        while True:
            try:
                dirpath, filename = self.rename(dirpath, tempfile, filename)
                logging.getLogger().debug('Renamed temporary file. %s', {'path': dirpath, 'temp': tempfile, 'name': filename})
                break
            except (FileExistsError, IsADirectoryError):
                logging.getLogger().debug('File exists. %s', {'path': dirpath, 'name': filename})
                version = version + 1
                filename = self.version(origin, version)

        filepath = os.path.join(dirpath, filename)
        logging.getLogger().info('Saved. %s', {'path': filepath})
        return filepath

    @staticmethod
    def version(filename, version):
        idx = filename.rfind('.')
        extension = ''
        if idx > 0:
            name = filename[:idx]
            extension = filename[idx:]
        else:
            name = filename
        return name + ' ' + '(' + str(version) + ')' + extension

    @staticmethod
    def write(filepath, handle):
        with open(filepath, 'w+b') as fd:
            if isinstance(handle, bytes):
                fd.write(handle)
            else:
                for chunk in handle.iter_content(chunk_size=8192):
                    fd.write(chunk)
        logging.getLogger().debug('Saved temporary file. %s', {'path': filepath})

    @staticmethod
    def get_local_file_info(local_file):
        path = Path(local_file)
        if not path.exists():
            logging.getLogger().error('The path %(local_file)s was not found', dict(local_file=local_file))
            raise LocalFileNotFound(local_file)
        if not path.is_file():
            logging.getLogger().error('The path %(local_file)s is not a file', dict(local_file=local_file))
            raise LocalFileNotFound(local_file)

        return dict(
            name=path.name,
            size=str(path.stat().st_size),
            mimetype=mimetypes.guess_type(local_file)
        )

    @staticmethod
    def compute_zip_file_name(cloud_directory, files):
        if len(files) > 1:
            path = Path(cloud_directory)
        else:
            path = Path(files[0])
        return path.stem + '.zip'

    def get_dirpath(self):
        dirpath = config.filesystem['dl']
        try:
            self.validate_directory(dirpath)
        except LocalDirectoryNotFound as error:
            dirpath = self.expanduser(dirpath)
            logging.getLogger().error('Download failed. Check the following directory exists. %s', {'path': dirpath})
            raise error
        return dirpath

    @staticmethod
    def join(*paths):
        return os.path.join(*paths)

    @staticmethod
    def split_file_directory(path):
        # Exists and file -> directory=parent, filename=name
        # Exists and dir -> directory=current filename=None
        # Not Exists and parent Exists -> directory=parent filename=name
        # Not Exists and parent not Exists -> Error
        path = os.path.expanduser(path)
        p = Path(path)
        if p.exists():
            if p.is_dir():
                filename = None
            else:
                filename = p.name
                p = p.parent
        elif p.parent.exists():
            filename = p.name
            p = p.parent
        else:
            raise LocalPathNotFound(path)
        return str(p.resolve()), filename

    @staticmethod
    def split_file_directory_with_defaults(path=None, default_filename=None):
        """
        Compute the destination file path.

        :param str path: A path to a file or a folder
        :param str default_filename: The default file name to use unless `path` argument specifies a file path

        :returns: A tuple including the destination directory and filename
        :rtype: (string, string)
        """
        directory = filename = None
        if path:
            directory, filename = FileSystem.split_file_directory(path)
        else:
            directory = FileSystem.instance().get_dirpath()

        if not filename:
            filename = default_filename

        return (directory, filename)



from ..client import NetworkHost, CTERAHost
from ..edge import decorator
from ..edge import connection
from ..edge import query
from ..edge import afp
from ..edge import aio
from ..edge import array
from ..edge import audit
from ..edge import backup
from ..edge import cache
from ..edge import cli
from ..edge import config
from ..edge import dedup
from ..edge import directoryservice
from ..edge import drive
from ..edge import ftp
from ..edge import groups
from ..edge import migration_tool
from ..edge import licenses
from ..edge import login
from ..edge import logs
from ..edge import mail
from ..edge import network
from ..edge import nfs
from ..edge import ntp
from ..edge import snmp
from ..edge import ssl
from ..edge import ssh
from ..edge import power
from ..edge import rsync
from ..edge import services
from ..edge import session
from ..edge import shares
from ..edge import shell
from ..edge import smb
from ..edge import support
from ..edge import sync
from ..edge import syslog
from ..edge import taskmgr
from ..edge import telnet
from ..edge import timezone
from ..edge import users
from ..edge import volumes
from ..edge import files
from ..edge import remote
from ..edge import uri
from ..edge import firmware


class Gateway(CTERAHost):  # pylint: disable=too-many-instance-attributes
    """
    Main class operating on a Gateway

    :ivar cterasdk.edge.config.Config config: Object holding the Gateway Configuration APIs
    :ivar cterasdk.edge.network.Network network: Object holding the Gateway Network APIs
    :ivar cterasdk.edge.licenses.Licenses licenses: Object holding the Gateway Licenses APIs
    :ivar cterasdk.edge.services.Services services: Object holding the Gateway Services APIs
    :ivar cterasdk.edge.directoryservice.DirectoryService directoryservice: Object holding the Gateway Active Directory APIs
    :ivar cterasdk.edge.telnet.Telnet telnet: Object holding the Gateway Telnet APIs
    :ivar cterasdk.edge.syslog.Syslog syslog: Object holding the Gateway Syslog APIs
    :ivar cterasdk.edge.taskmgr.Tasks tasks: Object holding the Gateway Background Tasks APIs
    :ivar cterasdk.edge.audit.Audit audit: Object holding the Gateway Audit APIs
    :ivar cterasdk.edge.mail.Mail mail: Object holding the Gateway Mail APIs
    :ivar cterasdk.edge.backup.Backup backup: Object holding the Gateway Backup APIs
    :ivar cterasdk.edge.sync.Sync sync: Object holding the Gateway Sync APIs
    :ivar cterasdk.edge.cache.Cache cache: Object holding the Gateway Cache APIs
    :ivar cterasdk.edge.snmp.SNMP snmp: Object holding the Gateway SNMP APIs
    :ivar cterasdk.edge.ssl.SSL ssl: Object holding the Gateway SSL APIs
    :ivar cterasdk.edge.ssh.SSH ssh: Object holding the Gateway SSH APIs
    :ivar cterasdk.edge.power.Power power: Object holding the Gateway Power APIs
    :ivar cterasdk.edge.users.Users users: Object holding the Gateway Users APIs
    :ivar cterasdk.edge.groups.Groups groups: Object holding the Gateway Groups APIs
    :ivar cterasdk.edge.migration_tool.MigrationTool mtool: Object holding the Edge Filer's Migration Tool APIs
    :ivar cterasdk.edge.drive.Drive drive: Object holding the Gateway Drive APIs
    :ivar cterasdk.edge.volumes.Volumes volumes: Object holding the Gateway Volumes APIs
    :ivar cterasdk.edge.array.Array array: Object holding the Gateway Array APIs
    :ivar cterasdk.edge.shares.Shares shares: Object holding the Gateway Shares APIs
    :ivar cterasdk.edge.smb.SMB smb: Object holding the Gateway SMB APIs
    :ivar cterasdk.edge.aio.AIO aio: Object holding the Gateway AIO APIs
    :ivar cterasdk.edge.ftp.FTP ftp: Object holding the Gateway FTP APIs
    :ivar cterasdk.edge.afp.AFP afp: Object holding the Gateway AFP APIs
    :ivar cterasdk.edge.nfs.NFS nfs: Object holding the Gateway NFS APIs
    :ivar cterasdk.edge.rsync.RSync rsync: Object holding the Gateway RSync APIs
    :ivar cterasdk.edge.timezone.Timezone timezone: Object holding the Gateway Timezone APIs
    :ivar cterasdk.edge.logs.Logs logs: Object holding the Gateway Logs APIs
    :ivar cterasdk.edge.ntp.NTP ntp: Object holding the Gateway NTP APIs
    :ivar cterasdk.edge.shell.Shell shell: Object holding the Gateway Shell APIs
    :ivar cterasdk.edge.cli.CLI cli: Object holding the Gateway CLI APIs
    :ivar cterasdk.edge.dedup.Dedup dedup: Object holding the Gateway Local Deduplication APIs
    :ivar cterasdk.edge.support.Support support: Object holding the Gateway Support APIs
    :ivar cterasdk.edge.files.FileBrowser files: Object holding the Gateway File Browsing APIs
    :ivar cterasdk.edge.firmware.Firmware firmware: Object holding the Gateway Firmware APIs
    """

    def __init__(self, host, port=None, https=False, Portal=None):
        """
        :param str host: The fully qualified domain name, hostname or an IPv4 address of the Gateway
        :param int,optional port: Set a custom port number (0 - 65535), If not set defaults to 80 for http and 443 for https
        :param bool,optional https: Set to True to require HTTPS, defaults to False
        :param cterasdk.object.Portal.Portal,optional Portal: The portal throught which the remote session was created, defaults to None
        """
        super().__init__(host, port, https)
        self._remote_access = False
        self._session = session.Session(self.host())
        if Portal is not None:
            self._Portal = Portal
            self._ctera_client = Portal._ctera_client
            self._session.start_remote_session(self._Portal.session())
        self.config = config.Config(self)
        self.dedup = dedup.Dedup(self)
        self.network = network.Network(self)
        self.licenses = licenses.Licenses(self)
        self.services = services.Services(self)
        self.directoryservice = directoryservice.DirectoryService(self)
        self.telnet = telnet.Telnet(self)
        self.syslog = syslog.Syslog(self)
        self.audit = audit.Audit(self)
        self.mail = mail.Mail(self)
        self.backup = backup.Backup(self)
        self.sync = sync.Sync(self)
        self.cache = cache.Cache(self)
        self.snmp = snmp.SNMP(self)
        self.ssl = ssl.SSL(self)
        self.ssh = ssh.SSH(self)
        self.power = power.Power(self)
        self.users = users.Users(self)
        self.groups = groups.Groups(self)
        self.drive = drive.Drive(self)
        self.volumes = volumes.Volumes(self)
        self.array = array.Array(self)
        self.shares = shares.Shares(self)
        self.smb = smb.SMB(self)
        self.aio = aio.AIO(self)
        self.ftp = ftp.FTP(self)
        self.afp = afp.AFP(self)
        self.nfs = nfs.NFS(self)
        self.rsync = rsync.RSync(self)
        self.timezone = timezone.Timezone(self)
        self.logs = logs.Logs(self)
        self.ntp = ntp.NTP(self)
        self.shell = shell.Shell(self)
        self.cli = cli.CLI(self)
        self.support = support.Support(self)
        self.files = files.FileBrowser(self)
        self.firmware = firmware.Firmware(self)
        self.tasks = taskmgr.Tasks(self)
        self.mtool = migration_tool.MigrationTool(self)

    @property
    def base_api_url(self):
        return uri.api(self)

    @property
    def base_file_url(self):
        return uri.files(self)

    @property
    def _session_id_key(self):
        return '_cteraSessionId_'

    @staticmethod
    def make_local_files_dir(full_path):
        return f'localFiles/{full_path}'

    @property
    def _omit_fields(self):
        return super()._omit_fields + [
            'config',
            'network',
            'licenses',
            'services',
            'directoryservice',
            'telnet',
            'syslog',
            'audit',
            'mail',
            'backup',
            'sync',
            'cache',
            'snmp',
            'ssl',
            'ssh',
            'power',
            'users',
            'groups',
            'drive',
            'volumes',
            'array',
            'shares',
            'smb',
            'aio',
            'ftp',
            'afp',
            'nfs',
            'rsync',
            'timezone',
            'logs',
            'ntp',
            'shell',
            'cli',
            'support',
            'files',
            'firmware',
            'tasks',
            'dedup',
            'mtool'
            ]

    @property
    def _login_object(self):
        return login.Login(self)

    @property
    def initialized(self):
        return not login.Login(self).info().isfirstlogin

    def _is_authenticated(self, function, *args, **kwargs):

        def is_nosession(path):
            return path.startswith('/nosession')

        def is_migration_auth(path):
            return path.startswith('/migration/rest/v1/auth/user')
        current_session = self.session()
        return current_session.authenticated() or current_session.initializing() or is_nosession(args[0]) or is_migration_auth(args[0])

    def test(self):
        """ Verification check to ensure the target host is a Gateway. """
        return connection.test(self)

    def remote_access(self):
        remote.remote_access(self, self._Portal)

    @decorator.authenticated
    def query(self, path, key, value):
        return query.query(self, path, key, value)

    @decorator.authenticated
    def show_query(self, path, key, value):
        query.show(self, path, key, value)

    @decorator.authenticated
    def rm(self, path):
        return super().delete(path, use_file_url=True)

    def _api(self):
        return uri.api(self)

    def _files(self):
        return uri.files(self)

    def _baseurl(self):
        return NetworkHost.baseurl(self)



import os
import ssl

from ..lib import TempfileServices


class CertificateServices:

    @staticmethod
    def add_trusted_cert(host, port):
        filepath = CertificateServices.save_cert_from_server(host, port)
        context = ssl.SSLContext(ssl.PROTOCOL_TLS)
        context.load_verify_locations(filepath)
        return context

    @staticmethod
    def save_cert_from_server(host, port):
        fd, filepath = TempfileServices.mkfile(host, '.pem')
        pem = ssl.get_server_certificate((host, port))
        os.write(fd, pem.encode())
        os.close(fd)
        return filepath



import re
import logging

from ..lib.task_manager_base import TaskBase
from ..exception import InputError
from .base_command import BaseCommand


class Task(TaskBase):

    def _get_task_id(self, ref):
        match = re.search('servers/[^/]*/bgTasks/[1-9][0-9]*$', ref)
        if not match:
            logging.getLogger().error('Invalid task id. %s', {'ref': ref})
            raise InputError('Invalid task id', ref, ['servers/server/bgTasks/107781'])
        return match.group(0)

    def get_task_status(self):
        if self.CTERAHost.session().in_tenant_context():
            return self.CTERAHost.execute('', 'getTaskStatus', self.path)
        return self.CTERAHost.get('/' + self.path)


class Tasks(BaseCommand):
    """ Portal Background Task APIs """

    def status(self, ref):
        """
        Get background task status

        :param str ref: Task reference
        """
        task = Task(self._portal, ref)
        return task.status()

    def wait(self, ref, retries=100, seconds=1):
        """
        Wait for background task to complete

        :param str ref: Task reference
        :param int,optional retries: Number of retries when sampling the task status, defaults to 100
        :param int,optional seconds: Number of seconds to wait between retries, defaults to 1
        """
        task = Task(self._portal, ref, retries, seconds)
        return task.wait()



from .base_command import BaseCommand
from ..common import Object


class Firmwares(BaseCommand):
    """
    Portal Firmware Repository APIs
    """
    def list_images(self):
        """
        List Firmware Images.\n
        """
        if self.session().in_tenant_context():
            param = Object()
            param._classname = 'FirmwareParam'  # pylint: disable=protected-access
            return self._portal.execute('', 'getFirmwares', param)
        return self._portal.get('/firmwares')



import logging
from zipfile import ZipFile

from .base_command import BaseCommand
from ..lib import FileSystem, X509Certificate, PrivateKey, TempfileServices, create_certificate_chain


class SSL(BaseCommand):
    """
    Portal SSL Certificate APIs
    """

    def __init__(self, portal):
        super().__init__(portal)
        self._filesystem = FileSystem.instance()

    def get(self):
        """
        Retrieve details of the current installed SSL certificate

        :return cterasdk.common.object.Object: An object including the SSL certificate details
        """
        logging.getLogger().info('Retrieving SSL certificate')
        response = self._portal.get('/settings/ca')
        logging.getLogger().info('Retrieved SSL certificate')
        return response

    @property
    def thumbprint(self):
        """
        Get the SHA1 thumbprint of the Portal SSL certificate
        """
        return self.get().thumbprint

    def export(self, destination=None):
        """
        Export the Portal SSL Certificate to a ZIP archive

        :param str,optional destination:
         File destination, defaults to the default directory
        """
        directory, filename = self._filesystem.split_file_directory_with_defaults(destination, 'certificate.zip')
        logging.getLogger().info('Exporting SSL certificate.')
        handle = self._portal.openfile('/admin/preview/exportCertificate', use_file_url=True)
        filepath = self._filesystem.save(directory, filename, handle)
        logging.getLogger().info('Exported SSL certificate. %s', {'filepath': filepath})
        return filepath

    def create_zip_archive(self, private_key, *certificates):
        """
        Create a ZIP archive that can be imported to CTERA Portal

        :param str private_key: The PEM-encoded private key, or a path to the PEM-encoded private key file
        :param list[str] certificates: The PEM-encoded certificates, or a list of paths of the PEM-encoded certificate files
        """
        tempdir = TempfileServices.mkdir()

        key_basename = 'private.key'
        key_object = PrivateKey.load_private_key(private_key)
        key_filepath = FileSystem.join(tempdir, key_basename)
        self._filesystem.write(key_filepath, key_object.pem_data)

        cert_basename = 'certificate'
        certificates = [X509Certificate.load_certificate(certificate) for certificate in certificates]
        certificate_chain = create_certificate_chain(*certificates)

        certificate_chain_zip_archive = None
        if certificate_chain:
            certificate_chain_zip_archive = FileSystem.join(tempdir, f'{cert_basename}.zip')
            with ZipFile(certificate_chain_zip_archive, 'w') as zip_archive:
                zip_archive.write(key_filepath, key_basename)
                for idx, certificate in enumerate(certificate_chain):
                    filename = f'{cert_basename}{idx if idx > 0 else ""}.crt'
                    filepath = FileSystem.join(tempdir, filename)
                    self._filesystem.write(filepath, certificate.pem_data)
                    zip_archive.write(filepath, filename)

        return certificate_chain_zip_archive

    def import_from_zip(self, zipfile):
        """
        Import an SSL Certificate to CTERA Portal from a ZIP archive

        :param str zipfile: A zip archive including the private key and SSL certificate chain
        """
        return self._import_certificate(zipfile)

    def import_from_chain(self, private_key, *certificates):
        """
        Import an SSL Certificate to CTERA Portal from a chain

        :param str private_key: The PEM-encoded private key, or a path to the PEM-encoded private key file
        :param list[str] certificates: The PEM-encoded certificates, or a list of paths of the PEM-encoded certificate files
        """
        zipflie = self.create_zip_archive(private_key, *certificates)
        return self.import_from_zip(zipflie)

    def _import_certificate(self, zipfile):
        info = self._filesystem.get_local_file_info(zipfile)
        logging.getLogger().info('Uploading SSL certificate.')
        with open(zipfile, 'rb') as fd:
            response = self._portal.upload(
                '/settings/importCertificate',
                dict(
                    name='upload',
                    certificate=(info['name'], fd, info['mimetype'][0])
                )
            )
            if not isinstance(response, str):
                logging.getLogger().error('Failed uploading SSL certificate. %s', {'reason': response.msg})
            logging.getLogger().info('Uploaded SSL certificate.')
            self._portal.startup.wait()



import logging

from ..common import Object, union
from . import query
from .enum import ICAPServices
from .base_command import BaseCommand


class Antivirus(BaseCommand):
    """
    Portal Antivirus APIs

    :ivar cterasdk.core.antivirus.AntivirusServers servers: Object holding the Portal antivirus server APIs
    """

    default = ['name', 'type']

    def __init__(self, portal):
        super().__init__(portal)
        self.servers = AntivirusServers(self._portal)

    def list_servers(self, include=None):
        """
        List the antivirus servers

        :param list[str],optional include: List of fields to retrieve, defaults to ``['name', 'type']``
        """
        include = union(include or [], Antivirus.default)
        param = query.QueryParamBuilder().include(include).build()
        return query.iterator(self._portal, '/antiviruses', param)

    def rescan(self):
        """
        Scan all files using the latest antivirus update. This may take a while
        """
        logging.getLogger().info("Starting antivirus rescan.")
        self._portal.execute('/servers', 'resetAllAVBG')
        logging.getLogger().info("Started antivirus rescan.")

    def suspend(self):
        """
        Suspend antivirus scanning
        """
        logging.getLogger().info("Suspending antivirus scanning.")
        self._portal.put('/settings/cloudFSSettings/antivirusSettings/isEnabled', False)
        logging.getLogger().info("Suspended antivirus scanning.")

    def unsuspend(self):
        """
        Unsuspend antivirus scanning
        """
        logging.getLogger().info("Unsuspending antivirus scanning.")
        self._portal.put('/settings/cloudFSSettings/antivirusSettings/isEnabled', True)
        logging.getLogger().info("Unsuspended antivirus scanning.")

    def status(self):
        """
        Get antivirus service status
        """
        param = Object()
        param.icapService = ICAPServices.Antivirus
        return self._portal.execute('', 'getIcapGlobalStatus', param)


class AntivirusServers(BaseCommand):

    def get(self, name):
        """
        Get an antivirus server's configuration

        :param str name: Server name
        """
        return self._portal.get(f'/antiviruses/{name}')

    def add(self, name, vendor, url, connection_timeout=5):
        """
        Add an antivirus server

        :param str name: Server name
        :param cterasdk.core.enum.AntivirusType vendor: Server type
        :param str url: Server URL (example: ``http://your-antivirus.server.local:1234/signature``)
        :param int,optional connection_timeout: Server connection timeout (in seconds), defaults to 5 seconds
        """
        param = Object()
        param._classname = 'Antivirus'  # pylint: disable=protected-access
        param.name = name
        param.type = vendor
        param.serverUrl = url
        param.connectionTimeoutSeconds = connection_timeout
        logging.getLogger().info("Adding antivirus server. %s", {'name': name, 'type': vendor, 'url': url})
        response = self._portal.add('/antiviruses', param)
        logging.getLogger().info("Added antivirus server. %s", {'name': name, 'type': vendor, 'url': url})
        return response

    def delete(self, name):
        """
        Remove an antivirus server
        """
        return self._portal.delete(f'/antiviruses/{name}')

    def suspend(self, name):
        """
        Suspend an antivirus server
        """
        logging.getLogger().info("Suspending antivirus server. %s", {'name': name})
        self._portal.put(f'/antiviruses/{name}/enabled', False)
        logging.getLogger().info("Suspended antivirus server. %s", {'name': name})

    def unsuspend(self, name):
        """
        Unsuspend antivirus scanning
        """
        logging.getLogger().info("Unsuspending antivirus server. %s", {'name': name})
        self._portal.put(f'/antiviruses/{name}/enabled', True)
        logging.getLogger().info("Unsuspended antivirus server. %s", {'name': name})



import logging
from .base_command import BaseCommand
from ..exception import CTERAException, InputError, ObjectNotFoundException
from .enum import PlanItem, PlanRetention
from ..common import union, convert_size, DataUnit, PolicyRuleConverter
from . import query


class Plans(BaseCommand):
    """
    Portal Plan APIs

    :ivar cterasdk.core.plans.PlanAutoAssignPolicy auto_assign: Object holding the Portal subscription plan auto assignment rules APIs
    """
    default = ['name']
    _allowed_storage_size_units = [DataUnit.GB, DataUnit.TB, DataUnit.PB]

    def __init__(self, portal):
        super().__init__(portal)
        self.auto_assign = PlanAutoAssignPolicy(self._portal)

    def _get_entire_object(self, name):
        """
        Get a subscription plan

        :param str name: Name of the subscription plan
        """
        try:
            return self._portal.get('/plans/' + name)
        except CTERAException as error:
            raise CTERAException('Could not find subscription plan', error)

    def by_name(self, names, include=None):
        """
        Get Plans by their names

        :param list[str],optional names: List of names of plans
        :param list[str],optional include: List of fields to retrieve, defaults to ['name']
        :param list[cterasdk.core.query.FilterBuilder],optional filters: List of additional filters, defaults to None

        :return: Iterator for all matching Plans
        :rtype: cterasdk.lib.iterator.Iterator
        """
        filters = [query.FilterBuilder('name').eq(name) for name in names]
        return self.list_plans(include, filters)

    def list_plans(self, include=None, filters=None):
        """
        List Plans

        :param list[str],optional include: List of fields to retrieve, defaults to ['name']
        :param list[],optional filters: List of additional filters, defaults to None

        :return: Iterator for all matching Plans
        :rtype: cterasdk.lib.iterator.Iterator
        """
        include = union(include or [], Plans.default)
        builder = query.QueryParamBuilder().include(include)
        filters = filters or []
        for query_filter in filters:
            builder.addFilter(query_filter)
        builder.orFilter((len(filters) > 1))
        param = builder.build()
        iterator = query.iterator(self._portal, '/plans', param)
        for plan in iterator:
            yield plan

    def get(self, name, include=None):
        """
        Retrieve subscription plan properties

        :param str name: Name of the subscription plan
        :param list[str] include: List of fields to retrieve, defaults to ['name']
        :return: The subscription plan, including the requested fields
        """
        include = union(include or [], Plans.default)
        include = ['/' + attr for attr in include]
        plan = self._portal.get_multi('/plans/' + name, include)
        if plan.name is None:
            raise ObjectNotFoundException('Could not find subscription plan', f'/plans/{name}', name=name)
        return plan

    def add(self, name, retention=None, quotas=None):
        """
        Add a subscription plan

        :param dict,optional retention: The data retention policy
        :param dict,optional quotas: The items included in the plan and their respective quota
        """
        plan = self._portal.default_class('Plan')
        plan.name = name
        Plans._assign_retention(plan, retention)
        Plans._assign_quotas(plan, quotas)
        try:
            response = self._portal.add('/plans', plan)
            logging.getLogger().info("Plan created. %s", {'plan': name})
            return response
        except CTERAException as error:
            logging.getLogger().error("Plan creation failed.")
            raise CTERAException('Plan creation failed', error)

    def modify(self, name, retention=None, quotas=None, apply_changes=True):
        """
        Modify a subscription plan

        :param dict,optional retention: The data retention policy
        :param dict,optional quotas: The items included in the plan and their respective quota
        :param bool,optional apply_changes: Apply provisioning changes immediately
        """
        plan = self._get_entire_object(name)
        Plans._assign_retention(plan, retention)
        Plans._assign_quotas(plan, quotas)
        try:
            response = self._portal.put('/plans/' + name, plan)
            logging.getLogger().info("Plan modified. %s", {'plan': name})
            if apply_changes:
                if self._portal.session().in_tenant_context():
                    self._portal.users.apply_changes(True)
                else:
                    self._portal.portals.apply_changes(True)
            return response
        except CTERAException as error:
            logging.getLogger().error("Could not modify subscription plan.")
            raise CTERAException('Could not modify subscription plan', error)

    @staticmethod
    def _assign_retention(plan, retention):
        if retention is not None:
            plan.retentionPolicy.retainAll = retention.get(PlanRetention.All, plan.retentionPolicy.retainAll)
            plan.retentionPolicy.hourly = retention.get(PlanRetention.Hourly, plan.retentionPolicy.hourly)
            plan.retentionPolicy.daily = retention.get(PlanRetention.Daily, plan.retentionPolicy.daily)
            plan.retentionPolicy.weekly = retention.get(PlanRetention.Weekly, plan.retentionPolicy.weekly)
            plan.retentionPolicy.monthly = retention.get(PlanRetention.Monthly, plan.retentionPolicy.monthly)
            plan.retentionPolicy.quarterly = retention.get(PlanRetention.Quarterly, plan.retentionPolicy.quarterly)
            plan.retentionPolicy.yearly = retention.get(PlanRetention.Yearly, plan.retentionPolicy.yearly)
            plan.retentionPolicy.retainDeleted = retention.get(PlanRetention.Deleted, plan.retentionPolicy.retainDeleted)

    @staticmethod
    def _assign_quotas(plan, quotas):
        if quotas is not None:
            plan.vGateways4.amount = quotas.get(PlanItem.EV4, plan.vGateways4.amount)
            plan.vGateways8.amount = quotas.get(PlanItem.EV8, plan.vGateways8.amount)
            plan.appliances.amount = quotas.get(PlanItem.EV16, plan.appliances.amount)  # EV16
            plan.vGateways32.amount = quotas.get(PlanItem.EV32, plan.vGateways32.amount)
            plan.vGateways64.amount = quotas.get(PlanItem.EV64, plan.vGateways64.amount)
            plan.vGateways128.amount = quotas.get(PlanItem.EV128, plan.vGateways128.amount)
            plan.workstationAgents.amount = quotas.get(PlanItem.WA, plan.workstationAgents.amount)
            plan.serverAgents.amount = quotas.get(PlanItem.SA, plan.serverAgents.amount)
            plan.cloudDrives.amount = quotas.get(PlanItem.Share, plan.cloudDrives.amount)
            plan.cloudDrivesLite.amount = quotas.get(PlanItem.Connect, plan.cloudDrivesLite.amount)
            plan.storage.amount = Plans._get_storage_amount(quotas.get(PlanItem.Storage), plan.storage.amount)

    @staticmethod
    def _get_storage_amount(new_amount, existing_amount):
        if new_amount is None:
            return existing_amount

        unit = DataUnit.GB
        if isinstance(new_amount, int):
            value = new_amount
        elif isinstance(new_amount, str):
            try:
                value = int(new_amount)
            except ValueError:
                value = int(new_amount[:-2])
                unit = new_amount[-2:]
        else:
            raise InputError("Invalid storage amount type", new_amount, ['int', 'str'])

        if unit not in Plans._allowed_storage_size_units:
            raise InputError("Invalid unit type", unit, Plans._allowed_storage_size_units)

        return convert_size(value, unit, DataUnit.B)

    def delete(self, name):
        """
        Delete a subscription plan

        :param str username: The name of the subscription plan
        """
        try:
            response = self._portal.delete('/plans/' + name)
            logging.getLogger().info("Plan deleted. %s", {'name': name})
            return response
        except CTERAException as error:
            logging.getLogger().error("Plan deletion failed.")
            raise CTERAException('Plan deletion failed', error)


class PlanAutoAssignPolicy(BaseCommand):

    def get_policy(self):
        """
        Get plans auto assignment policy
        """
        return self._portal.execute('', 'getPlanAutoAssignmentRules')

    def set_policy(self, rules, apply_default=None, default=None, apply_changes=True):
        """
        Set plans auto assignment policy

        :param list[cterasdk.common.types.PolicyRule] rules: List of policy rules
        :param bool,optional apply_default: If no match found, apply default plan. If not passed, the current config will be kept
        :param str,optional default: Name of a plan to assign if no match found. Ignored unless the ``apply_default`` is set to ``True``
        :param bool,optional apply_changes: Apply provisioning changes upon update, defaults to ``True``
        """
        plans = {rule.assignment for rule in rules}
        if default:
            plans.add(default)
        plans = list(plans)
        portal_plans = {plan.name: plan for plan in self._portal.plans.by_name(plans, ['baseObjectRef'])}

        not_found = [plan for plan in plans if plan not in portal_plans.keys()]
        if not_found:
            logging.getLogger().error('Could not find one or more plans. %s', {'plans': not_found})
            raise CTERAException('Could not find one or more plans', None, plans=not_found)

        policy = self.get_policy()

        if apply_default is False:
            policy.defaultPlan = None
        elif apply_default is True and default:
            policy.defaultPlan = portal_plans.get(default).baseObjectRef

        policy_rules = [PolicyRuleConverter.convert(rule, 'PlanAutoAssignmentRule', 'plan',
                        portal_plans.get(rule.assignment).baseObjectRef) for rule in rules]
        policy.planAutoAssignmentRules = policy_rules

        response = self._portal.execute('', 'setPlanAutoAssignmentRules', policy)
        logging.getLogger().info('Set plans auto assignment rules.')

        if apply_changes:
            self._portal.users.apply_changes(True)

        return response



import logging
from datetime import datetime

from .base_command import BaseCommand
from ..lib import Iterator, Command
from ..core import enum
from ..common import Object
from . import query


class Logs(BaseCommand):
    """
    Portal Logs APIs

    :ivar cterasdk.core.logs.Alerts alerts: Object holding the Portal Log Based Alerts APIs
    """

    def __init__(self, portal):
        super().__init__(portal)
        self.alerts = Alerts(self._portal)

    def device(self,
               name,
               topic=enum.LogTopic.System,
               min_severity=enum.Severity.INFO,
               before=None,
               after=None,
               filters=None):
        """
        Get device logs from the Portal

        :param str name: Name of a device
        :param cterasdk.core.enum.LogTopic,optional topic: Log topic to get, defaults to cterasdk.core.enum.LogTopic.System
        :param cterasdk.core.enum.Severity,optional min_severity:
         Minimun severity of logs to get, defaults to cterasdk.core.enum.Severity.INFO
        :param str,optional before: Get logs before this date (in format "%m/%d/%Y %H:%M:%S"), defaults to None
        :param str,optional after: Get logs after this date (in format "%m/%d/%Y %H:%M:%S"), defaults to None
        :param list[cterasdk.core.query.FilterBuilder],optional filters: List of additional filters, defaults to None

        :return: Iterator for all matching logs
        :rtype: cterasdk.lib.iterator.Iterator[cterasdk.object.Object]
        """
        return self.get(topic, min_severity, enum.OriginType.Device, name, before, after, filters)

    def get(self, topic=enum.LogTopic.System, min_severity=enum.Severity.INFO, origin_type=enum.OriginType.Portal, origin=None,
            before=None, after=None, filters=None):
        """
        Get logs from the Portal

        :param cterasdk.core.enum.LogTopic,optional topic: Log topic to get, defaults to cterasdk.core.enum.LogTopic.System
        :param cterasdk.core.enum.Severity,optional min_severity:
         Minimun severity of logs to get, defaults to cterasdk.core.enum.Severity.INFO
        :param cterasdk.core.enum.OriginType,optional origin_type:
         Origin type of the logs to get, defaults to cterasdk.core.enum.OriginType.Portal
        :param str,optional origin: Log origin (e.g. device name, Portal server name), defaults to None
        :param str,optional before: Get logs before this date (in format "%m/%d/%Y %H:%M:%S"), defaults to None
        :param str,optional after: Get logs after this date (in format "%m/%d/%Y %H:%M:%S"), defaults to None
        :param list[cterasdk.core.query.FilterBuilder],optional filters: List of additional filters, defaults to None

        :return: Iterator for all matching logs
        :rtype: cterasdk.lib.iterator.Iterator[cterasdk.object.Object]
        """
        builder = query.QueryParamBuilder().put('topic', topic).put('minSeverity', min_severity)

        builder.addFilter(query.FilterBuilder('originType').eq(origin_type))

        if before is not None:
            builder.addFilter(query.FilterBuilder('time').before(self._strptime(before)))

        if after is not None:
            builder.addFilter(query.FilterBuilder('time').after(self._strptime(after)))

        if origin is not None:
            builder.addFilter(query.FilterBuilder.ref('origin').eq(origin))

        if filters:
            for user_filter in filters:
                builder.addFilter(user_filter)

        param = builder.build()
        function = Command(self._query_logs)

        return Iterator(function, param)

    def _query_logs(self, param):
        response = self._portal.execute('', 'queryLogs', param)
        return (response.hasMore, response.logs)

    @staticmethod
    def _strptime(datetime_str):
        try:
            return datetime.strptime(datetime_str, '%m/%d/%Y %H:%M:%S')
        except ValueError as error:
            raise error


class Alerts(BaseCommand):
    """
    Portal Log Based Alerts APIs
    """

    def add(self, name, description=None, topic=None, log=None, min_severity=None, origin_type=None, content=None):
        """
        Add a Log Based Alert

        :param str name: Alert name
        :param str,optional description: Alert description
        :param cterasdk.core.enum.LogTopic,optional topic: Log topic to get, defaults to any topic
        :param str,optional log: Class name of the log
        :param cterasdk.core.enum.Severity,optional min_severity: Minimun severity for triggering an alert, defaults to any severity
        :param cterasdk.core.enum.OriginType,optional origin_type: Origin type of the log, defaults to any origin
        :param str content: Content of the log message
        :returns: A list of alerts
        :rtype: list[cterasdk.common.object.Object]
        """
        alert = Object()
        alert.id = name
        alert._classname = 'AlertRule'  # pylint: disable=protected-access
        if description:
            alert.description = description
        if log:
            alert.logName = log
        if content:
            alert.messageContent = content
        if min_severity:
            alert.minSeverity = min_severity
        if origin_type:
            alert.originType = origin_type
        if topic:
            alert.topic = topic
        alerts = self.get()
        alerts.append(alert)
        return self.put(alerts)

    def put(self, alerts):
        """
        Set Log Based Alerts
         Use :func:`cterasdk.core.types.AlertBuilder` to build log based alerts

        :param list[cterasdk.core.types.Alert] alerts: List of alerts
        """
        logging.getLogger().info('Updating log based alerts.')
        response = self._portal.put(self._context, alerts)
        logging.getLogger().info('Log based alerts updated.')
        return response

    def get(self):
        """
        Get a List of Log Based Alerts

        :returns: A list of alerts
        :rtype: list[cterasdk.common.object.Object]
        """
        return self._portal.get(self._context)

    def delete(self, name):
        """
        Remove a Log Based Alert

        :param str name: Alert name
        """
        alerts = [alert for alert in self.get() if alert.id != name]
        self.put(alerts)

    @property
    def _context(self):
        return f'{"" if self._portal.session().in_tenant_context() else "/settings"}/alerts'



def api(Portal):
    path = ''
    current_session = Portal.session()
    if current_session.authenticated() and current_session.is_local_auth() and current_session.in_tenant_context():
        path = '/portals/' + current_session.tenant()
    return Portal.base_portal_url + '/api' + path



import re
import logging

from ..common import Object
from ..exception import CTERAException, CTERAConnectionError
from .base_command import BaseCommand
from .types import TCPService


class DirectoryService(BaseCommand):
    """
    Gateway Active Directory configuration APIs
    """

    def connected(self):
        """
        Get the Active Directory join status
        """
        return self._gateway.get('/status/fileservices/cifs/joinStatus') == 0

    def connect(self, domain, username, password, ou=None, check_connection=False):
        """
        Connect the Gateway to an Active Directory

        :param str domain: The active directory domain to connect to
        :param str username: The user name to use when connecting to the active directory services
        :param str password: The password to use when connecting to the active directory services
        :param str,optional ou: The OU path to use when connecting to the active directory services, defaults to None
        :param bool,optional check_connection: Check connectivity before attempting to connect to directory services, defaults to `False`
        """
        if check_connection:
            self._check_domain_connectivity(domain)

        cifs = self._gateway.get('/config/fileservices/cifs')
        cifs.type = "domain"
        cifs.domain = domain
        cifs.workgroup = None
        self._gateway.put('/config/fileservices/cifs', cifs)

        param = Object()
        param.username = username
        param.password = password
        if ou is not None:
            param.ouPath = ou
        logging.getLogger().info("Connecting to Active Directory. %s", {'domain': domain, 'user': username})

        try:
            self._gateway.execute("/status/fileservices/cifs", "joinDomain", param)
        except CTERAException as error:
            logging.getLogger().error("Failed connecting to Active Directory.")
            raise error
        logging.getLogger().info("Connected to Active Directory.")

    def _check_domain_connectivity(self, domain):
        port = 389
        domain_controllers = self.get_static_domain_controller()
        domain_controllers = re.findall(r'\b(?:[0-9]{1,3}\.){3}[0-9]{1,3}\b', domain_controllers) if domain_controllers else [domain]
        connection_results = self._gateway.network.diagnose([TCPService(host, port) for host in domain_controllers])
        for connection_result in connection_results:
            if not connection_result.is_open:
                logging.getLogger().error("Connection failed. No traffic allowed over port %(port)s", dict(port=connection_result.port))
                raise CTERAConnectionError('Unable to establish connection', None, host=connection_result.host,
                                           port=connection_result.port, protocol='LDAP')

    def get_static_domain_controller(self):
        """
        Retrieve the static domain controller configuration

        :return: A FQDN, hostname or ip address of the domain controller
        :rtype: str
        """
        return self._gateway.get('/config/fileservices/cifs/passwordServer')

    def set_static_domain_controller(self, dc):
        """
        Configure the Gateway to use a static domain controller

        :param str dc: The FQDN, hostname or ip address of the domain controller
        :return: The FQDN, hostname or ip address of the domain controller
        :rtype: str
        """
        return self._gateway.put('/config/fileservices/cifs/passwordServer', dc)

    def remove_static_domain_controller(self):
        """
        Delete the static domain controller configuration
        """
        self._gateway.put('/config/fileservices/cifs/passwordServer', None)

    def get_advanced_mapping(self):
        """
        Retrieve directory services advanced mapping configuration

        :returns: A dictionary of domain mapping objects
        :rtype: dict
        """
        return {mapping.domainFlatName: mapping for mapping in self._gateway.get('/config/fileservices/cifs/idMapping/map')}

    def set_advanced_mapping(self, mappings):
        """
        Configure advanced mapping

        :param list[cterasdk.common.types.ADDomainIDMapping] mappings: List of domains and their UID/GID mapping range
        """
        if not self.connected():
            raise CTERAException('Failed to configure advanced mapping. Not connected to directory services.')

        domains = self.domains()
        advanced_mapping = self._gateway.get('/config/fileservices/cifs/idMapping/map')
        advanced_mapping = []
        for mapping in mappings:
            if mapping.domainFlatName in domains:
                advanced_mapping.append(mapping)
            else:
                logging.getLogger().warning('Invalid mapping. Could not find domain. %s', {'domain': mapping.domainFlatName})

        logging.getLogger().debug('Updating advanced mapping. %s', {
            'domains': [mapping.domainFlatName for mapping in advanced_mapping]
        })
        response = self._gateway.put('/config/fileservices/cifs/idMapping/map', advanced_mapping)
        logging.getLogger().info('Updated advanced mapping.')

        return response

    def get_connected_domain(self):
        """
        Get the connected domain information

        :return cterasdk.common.object.Object:
        """
        cifs = self._gateway.get('/config/fileservices/cifs')
        obj = Object()
        obj.type = cifs.type
        obj.domain = cifs.domain
        obj.workgroup = cifs.workgroup
        return obj

    def domains(self):
        """
        Get all domains

        :return list(str): List of names of all discovered domains
        """
        return [domain.flatName for domain in self._gateway.execute('/status/fileservices/cifs', 'enumDiscoveredDomains')]

    def disconnect(self):
        """
        Disconnect from Active Directory Service
        """
        logging.getLogger().info("Disconnecting from Active Directory.")

        cifs = self._gateway.get('/config/fileservices/cifs')
        cifs.type = "workgroup"
        cifs.workgroup = "CTERA"
        cifs.domain = None

        self._gateway.put('/config/fileservices/cifs', cifs)
        logging.getLogger().info("Disconnected from Active Directory.")



import logging


def openfile(ctera_host, path):
    fullpath = path.fullpath()
    logging.getLogger().info('Obtaining file handle. %s', {'path': fullpath})
    return ctera_host.openfile(ctera_host.make_local_files_dir(fullpath), use_file_url=True)



import logging

from ..common import Object
from .base_command import BaseCommand


class Mail(BaseCommand):
    """ Gateway Mail Server configuration APIs """

    def enable(self, smtp_server, port=25, username=None, password=None, use_tls=True):
        """
        Enable e-mail delivery using a custom SMTP server

        :param str smtp_server: Address of the SMTP Server
        :param int,optional port: The listening port of the SMTP Server, defaults to 25
        :param str,optional username: The user name of the SMTP Server, defaults to None
        :param str,optional password: The password of the SMTP Server, defaults to None
        :param bool,optional use_tls: Use TLS when connecting to the SMTP Server, defaults to True
        """
        settings = self._gateway.get('/config/logging/alert')
        settings.useCustomServer = True
        settings.SMTPServer = smtp_server

        if settings.port != port:
            settings.port = port

        if username is not None and password is not None:
            settings.useAuth = True
            settings.auth = Object()
            settings.auth.username = username
            settings.auth.password = password

        if settings.useTLS != use_tls:
            settings.useTLS = use_tls

        logging.getLogger().info('Enabling mail server.')

        self._gateway.put('/config/logging/alert', settings)

        logging.getLogger().info(
            'Updated mail server settings. %s',
            {'SMTPServer': smtp_server, 'port': port, 'username': username, 'useTLS': use_tls}
        )

    def disable(self):
        """ Disable e-mail delivery using a custom SMTP server """
        logging.getLogger().info('Disabling mail server.')
        self._gateway.put('/config/logging/alert/useCustomServer', False)
        logging.getLogger().info('Mail server disabled.')



import logging

from .enum import Mode
from ..exception import CTERAException
from .base_command import BaseCommand


class RSync(BaseCommand):
    """ Gateway RSync configuration """

    def get_configuration(self):
        """
        Get the current RSync configuration

        :return cterasdk.common.object.Object:
        """
        return self._gateway.get('/config/fileservices/rsync')

    def enable(self):
        """ Enable FTP """
        self._set_mode(True)

    def disable(self):
        """ Disable FTP """
        self._set_mode(False)

    def is_disabled(self):
        """ Check if the Rsync server is disabled """
        return self._gateway.get('/config/fileservices/rsync/server') == Mode.Disabled

    def _set_mode(self, enabled):
        """ Disable RSync """
        logging.getLogger().info('%s RSync server.', ('Enabling' if enabled else 'Disabling'))
        self._gateway.put('/config/fileservices/rsync/server', Mode.Enabled if enabled else Mode.Disabled)
        logging.getLogger().info('RSync server %s.', ('enabled' if enabled else 'disabled'))

    def modify(
            self,
            port=None,
            max_connections=None):
        """
        Modify the RSync Configuration. Parameters that are not passed will not be affected

        :param int,optional port: RSync Port
        :param int,optional max_connections: Maximum Connections
        """
        config = self.get_configuration()
        if config.server != Mode.Enabled:
            raise CTERAException("RSync must be enabled in order to modify its configuration")
        if port is not None:
            config.port = port
        if max_connections is not None:
            config.maxConnections = max_connections
        self._gateway.put('/config/fileservices/rsync', config)
